"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from copy import deepcopy
from typing import Tuple, Any, List, Iterable, Optional

import numpy
import torch
import torchvision.transforms.functional as TTF
from PIL import Image
from scipy.ndimage.filters import gaussian_filter
from scipy.ndimage.interpolation import map_coordinates
from torch import Tensor

from fuse.utils.utils_param_sampler import FuseUtilsParamSamplerGaussianPatch as Gaussian
from fuse.utils.utils_param_sampler import FuseUtilsParamSamplerRandBool as RandBool
from fuse.utils.utils_param_sampler import FuseUtilsParamSamplerRandInt as RandInt
from fuse.utils.utils_param_sampler import FuseUtilsParamSamplerUniform as Uniform


######## Affine augmentation
def aug_op_affine(aug_input: Tensor, rotate: float = 0.0, translate: Tuple[float, float] = (0.0, 0.0),
                  scale: Tuple[float, float] = 1.0, flip: Tuple[bool, bool] = (False, False), shear: float = 0.0,
                  channels: Optional[List[int]] = None) -> Tensor:
    """
    Affine augmentation
    :param aug_input: 2D tensor representing an image to augment, shape [num_channels, height, width] or [height, width]
    :param rotate: angle [0.0 - 360.0]
    :param translate: translation per spatial axis (number of pixels). The sign used as the direction.
    :param scale: scale factor
    :param flip: flip per spatial axis flip[0] for vertical flip and flip[1] for horizontal flip
    :param shear: shear factor
    :param channels: apply the augmentation on the specified channels. Set to None to apply to all channels.
    :return: the augmented image
    """
    # Support for 2D inputs - implicit single channel
    if len(aug_input.shape) == 2:
        aug_input = aug_input.unsqueeze(dim=0)
        remember_to_squeeze = True
    else:
        remember_to_squeeze = False

    # convert to PIL (required by affine augmentation function)
    if channels is None:
        channels = list(range(aug_input.shape[0]))
    aug_tensor = aug_input
    for channel in channels:
        aug_channel_tensor = aug_input[channel].numpy()
        aug_channel_tensor = Image.fromarray(aug_channel_tensor)
        aug_channel_tensor = TTF.affine(aug_channel_tensor, angle=rotate, scale=scale, translate=translate, shear=shear)
        if flip[0]:
            aug_channel_tensor = TTF.vflip(aug_channel_tensor)
        if flip[1]:
            aug_channel_tensor = TTF.hflip(aug_channel_tensor)

        # convert back to torch tensor
        aug_channel_tensor = numpy.array(aug_channel_tensor)
        aug_channel_tensor = torch.from_numpy(aug_channel_tensor)

        # set the augmented channel
        aug_tensor[channel] = aug_channel_tensor

    # squeeze back to 2-dim if needed
    if remember_to_squeeze:
        aug_tensor = aug_tensor.squeeze(dim=0)

    return aug_tensor


def aug_op_affine_group(aug_input: Tuple[Tensor], **kwargs) -> Tuple[Tensor]:
    """
    Applies same augmentation on multiple tensors. For example, augmenting both input image and its corresponding
    segmentation mask in the same way. This method wraps 'aug_op_affine'.
    :param aug_input: tuple of tensors
    :param kwargs:    augmentation params, same kwargs as 'aug_op_affine' - see docstring there
    :return: tuple of tensors, all augmented the same way
    """
    return tuple((aug_op_affine(element, **kwargs) for element in aug_input))


def aug_op_crop_and_resize(aug_input: Tensor,
                           scale: Tuple[float, float],
                           channels: Optional[List[int]] = None) -> Tensor:
    """
    Alternative to rescaling: center crop and resize back to the original dimensions. if scale is bigger than 1.0. the image first padded.
    :param aug_input: The tensor to augment
    :param scale: tuple of positive floats
    :param channels: apply augmentation on the specified channels or None for all of them
    :return: the augmented tensor
    """
    if len(aug_input.shape) == 2:
        aug_input = aug_input.unsqueeze(dim=0)
        remember_to_squeeze = True
    else:
        remember_to_squeeze = False

    if channels is None:
        channels = list(range(aug_input.shape[0]))
    aug_tensor = aug_input
    for channel in channels:
        aug_channel_tensor = aug_input[channel]

        if scale[0] != 1.0 or scale[1] != 1.0:
            cropped_shape = (int(aug_channel_tensor.shape[0] * scale[0]), int(aug_channel_tensor.shape[1] * scale[1]))
            padding = [[0, 0], [0, 0]]
            for dim in range(2):
                if scale[dim] > 1.0:
                    padding[dim][0] = (cropped_shape[dim] - aug_channel_tensor.shape[dim]) // 2
                    padding[dim][1] = (cropped_shape[dim] - aug_channel_tensor.shape[dim]) - padding[dim][0]
            aug_channel_tensor_pad = TTF.pad(aug_channel_tensor.unsqueeze(0), (padding[1][0], padding[0][0], padding[1][1], padding[0][1]))
            aug_channel_tensor_cropped = TTF.center_crop(aug_channel_tensor_pad, cropped_shape)
            aug_channel_tensor = TTF.resize(aug_channel_tensor_cropped, aug_channel_tensor.shape).squeeze(0)
            # set the augmented channel
            aug_tensor[channel] = aug_channel_tensor

    # squeeze back to 2-dim if needed
    if remember_to_squeeze:
        aug_tensor = aug_tensor.squeeze(dim=0)

    return aug_tensor


######## Color augmentation
def aug_op_clip(aug_input: Tensor, clip: Tuple[float, float] = (-1.0, 1.0)) -> Tensor:
    """
    Clip pixel values
    :param aug_input: the tensor to clip
    :param clip: values for clipping from both sides
    :return: Clipped tensor
    """
    aug_tensor = aug_input
    if clip is not None:
        aug_tensor = torch.clamp(aug_tensor, clip[0], clip[1], out=aug_tensor)
    return aug_tensor


def aug_op_add_col(aug_input: Tensor, add: float) -> Tensor:
    """
    Adding a values to all pixels
    :param aug_input: the tensor to augment
    :param add: the value to add to each pixel
    :return: the augmented tensor
    """
    aug_tensor = aug_input + add
    aug_tensor = aug_op_clip(aug_tensor, clip=(0, 1))
    return aug_tensor


def aug_op_mul_col(aug_input: Tensor, mul: float) -> Tensor:
    """
    multiply each pixel
    :param aug_input: the tensor to augment
    :param mul: the multiplication factor
    :return: the augmented tensor
    """
    input_tensor = aug_input * mul
    input_tensor = aug_op_clip(input_tensor, clip=(0, 1))
    return input_tensor


def aug_op_gamma(aug_input: Tensor, gain: float, gamma: float) -> Tensor:
    """
    Gamma augmentation
    :param aug_input: the tensor to augment
    :param gain: gain factor
    :param gamma: gamma factor
    :return: None
    """
    input_tensor = (aug_input ** gamma) * gain
    input_tensor = aug_op_clip(input_tensor, clip=(0, 1))
    return input_tensor


def aug_op_contrast(aug_input: Tensor, factor: float) -> Tensor:
    """
    Adjust contrast (notice - calculated across the entire input tensor, even if it's 3d)
    :param aug_input:the tensor to augment
    :param factor: contrast factor.   1.0 is neutral
    :return: the augmented tensor
    """
    calculated_mean = aug_input.mean()
    input_tensor = ((aug_input - calculated_mean) * factor) + calculated_mean
    input_tensor = aug_op_clip(input_tensor, clip=(0, 1))
    return input_tensor


def aug_op_color(aug_input: Tensor, add: Optional[float] = None, mul: Optional[float] = None,
                 gamma: Optional[float] = None, contrast: Optional[float] = None, channels: Optional[List[int]] = None):
    """
    Color augmentaion: including addition, multiplication, gamma and contrast adjusting
    :param aug_input: the tensor to augment
    :param add: value to add to each pixel
    :param mul: multiplication factor
    :param gamma: gamma factor
    :param contrast: contrast factor
    :param channels: Apply clipping just over the specified channels. If set to None will apply on all channels.
    :return:
    """
    aug_tensor = aug_input
    if channels is None:
        if add is not None:
            aug_tensor = aug_op_add_col(aug_tensor, add)
        if mul is not None:
            aug_tensor = aug_op_mul_col(aug_tensor, mul)
        if gamma is not None:
            aug_tensor = aug_op_gamma(aug_tensor, 1.0, gamma)
        if contrast is not None:
            aug_tensor = aug_op_contrast(aug_tensor, contrast)
    else:
        if add is not None:
            aug_tensor[channels] = aug_op_add_col(aug_tensor[channels], add)
        if mul is not None:
            aug_tensor[channels] = aug_op_mul_col(aug_tensor[channels], mul)
        if gamma is not None:
            aug_tensor[channels] = aug_op_gamma(aug_tensor[channels], 1.0, gamma)
        if contrast is not None:
            aug_tensor[channels] = aug_op_contrast(aug_tensor[channels], contrast)

    return aug_tensor


######## Gaussian noise
def aug_op_gaussian(aug_input: Tensor, mean: float = 0.0, std: float = 0.03, channels: Optional[List[int]] = None) -> Tensor:
    """
    Add gaussian noise
    :param aug_input: the tensor to augment
    :param mean: mean gaussian distribution
    :param std:  std gaussian distribution
    :param channels: Apply just over the specified channels. If set to None will apply on all channels.
    :return: the augmented tensor
    """
    aug_tensor = aug_input
    if channels is None:
        rand_patch = Gaussian(aug_tensor.shape, mean, std).sample()
        aug_tensor = aug_tensor + rand_patch
    else:
        rand_patch = Gaussian(aug_tensor[channels].shape, mean, std).sample()
        aug_tensor[channels] = aug_tensor[channels] + rand_patch
    return aug_tensor


def aug_op_elastic_transform(aug_input: Tensor, alpha: float = 1, sigma: float = 50, channels: Optional[List[int]] = None):
    """Elastic deformation of images as described in [Simard2003]_.
    .. [Simard2003] Simard, Steinkraus and Platt, "Best Practices for
       Convolutional Neural Networks applied to Visual Document Analysis",
       :param aug_input: input tensor of shape (C,Y,X)
       :param alpha: global pixel shifting (correlated to the article)
       :param sigma: Gaussian filter parameter
       :param channels: which channels to apply the augmentation
       :return distorted image
    """
    random_state = numpy.random.RandomState(None)
    if channels is None:
        channels = list(range(aug_input.shape[0]))
    aug_tensor = aug_input.numpy()
    for channel in channels:
        aug_channel_tensor = aug_input[channel].numpy()
        shape = aug_channel_tensor.shape
        dx1 = gaussian_filter((random_state.rand(*shape) * 2 - 1), sigma, mode="constant", cval=0) * alpha
        dx2 = gaussian_filter((random_state.rand(*shape) * 2 - 1), sigma, mode="constant", cval=0) * alpha

        x1, x2 = numpy.meshgrid(numpy.arange(shape[0]), numpy.arange(shape[1]))
        indices = numpy.reshape(x2 + dx2, (-1, 1)), numpy.reshape(x1 + dx1, (-1, 1))

        distored_image = map_coordinates(aug_channel_tensor, indices, order=1, mode='reflect')
        distored_image = distored_image.reshape(aug_channel_tensor.shape)
        aug_tensor[channel] = distored_image
    return torch.from_numpy(aug_tensor)


######### Default / Example augmentation pipline for a 2D image
def aug_image_default_pipeline(input_pointer: str) -> List[Any]:
    """
    Return default image augmentation pipeline. optimised for breast project (GMP model).
    In case paramter tunning is required - copy and change the values
    :param input_pointer: global dict pointer to the image
    :return: the default pipeline
    """
    return [
        [
            (input_pointer,),
            aug_op_affine,
            {'rotate': Uniform(-30.0, 30.0), 'translate': (RandInt(-10, 10), RandInt(-10, 10)),
             'flip': (RandBool(0.3), RandBool(0.3)), 'scale': Uniform(0.9, 1.1)},
            {'apply': RandBool(0.5)}
        ],
        [
            (input_pointer,),
            aug_op_color,
            {'add': Uniform(-0.06, 0.06), 'mul': Uniform(0.95, 1.05), 'gamma': Uniform(0.9, 1.1),
             'contrast': Uniform(0.85, 1.15)},
            {'apply': RandBool(0.5)}
        ],
        [
            (input_pointer,),
            aug_op_gaussian,
            {'std': 0.03},
            {'apply': RandBool(0.5)}
        ],
    ]


# general utilities
def aug_pipeline_step_replicate(step: List, key: str, values: Iterable) -> List[List]:
    """
    Replicate a step, but set different value for each replication for the specified key
    :param step: The step to replicate
    :param key: the key to override (withing te augmentation dunction input)
    :param values: Iterable specify the value for each replication
    :return:
    """
    list_of_steps = []
    for value in values:
        step_copy = deepcopy(step)
        step_copy[2][key] = value
        list_of_steps.append(step_copy)

    return list_of_steps


def aug_op_rescale_pixel_values(aug_input: Tensor, target_range: Tuple[float, float] = (-1.0, 1.0)) -> Tensor:
    """
    Scales pixel values to specific range.
    :param aug_input:       input tensor
    :param target_range:    target range, (min, max)
    :return:  rescaled tensor
    """
    max_val = aug_input.max()
    min_val = aug_input.min()
    if min_val == max_val == 0:
        return aug_input
    aug_input = aug_input - min_val
    aug_input = aug_input / (max_val - min_val)
    aug_input = aug_input * (target_range[1] - target_range[0])
    aug_input = aug_input + target_range[0]
    return aug_input


def squeeze_3d_to_2d(aug_input: Tensor, axis_squeeze: str) -> Tensor:
    '''
    squeeze selected axis of volume image into channel dimension, in
    order to fit the 2D augmentation functions
    :param aug_input: input of shape: (channel, z, y, x)
    :return:
    '''
    # aug_input shape is [channels, z, y, x]
    if axis_squeeze == 'y':
        aug_input = aug_input.permute((0, 2, 1, 3))
        # aug_input shape is [channels, y, z, x]
    elif axis_squeeze == 'x':
        aug_input = aug_input.permute((0, 3, 2, 1))
        # aug_input shape is [channels, x, y, z]
    else:
        assert axis_squeeze == 'z', "axis squeeze must be a string of either x, y, or z"
    return aug_input.reshape((aug_input.shape[0] * aug_input.shape[1],) + aug_input.shape[2:])


def unsqueeze_2d_to_3d(aug_input: Tensor, channels: int, axis_squeeze: str) -> Tensor:
    '''
    unsqueeze selected axis to original shape, and add the batch dimension
    :param aug_input:
    :return:
    '''
    aug_input = aug_input
    aug_input = aug_input.reshape((channels, aug_input.shape[0] // channels) + aug_input.shape[1:])
    if axis_squeeze == 'y':
        aug_input = aug_input.permute((0, 2, 1, 3))
        # aug_input shape is [channels, z, y, x]
    elif axis_squeeze == 'x':
        aug_input = aug_input.permute((0, 3, 2, 1))
        # aug_input shape is [channels, z, y, x]
    else:
        assert axis_squeeze == 'z', "axis squeeze must be a string of either x, y, or z"
    return aug_input


def rotation_in_3d(aug_input: Tensor, z_rot: float = 0.0, y_rot: float = 0.0, x_rot: float = 0):
    """
    rotates an input tensor around an axis, when for example z_rot is chosen,
    the rotation is in the x-y plane.
    Note: rotation angles are in relation to the original axis (not the rotated one)
    rotation angles should be given in degrees
    :param aug_input:image input should be in shape [channel, z, y, x]
    :param z_rot: angle to rotate x-y plane clockwise
    :param y_rot: angle to rotate x-z plane clockwise
    :param x_rot: angle to rotate z-y plane clockwise
    :return:
    """
    assert len(aug_input.shape) == 4  # will only work for 3d
    channels = aug_input.shape[0]
    if z_rot != 0:
        squeez_img = squeeze_3d_to_2d(aug_input, axis_squeeze='z')
        rot_squeeze = aug_op_affine(squeez_img, rotate=z_rot)
        aug_input = unsqueeze_2d_to_3d(rot_squeeze, channels, 'z')
    if x_rot != 0:
        squeez_img = squeeze_3d_to_2d(aug_input, axis_squeeze='x')
        rot_squeeze = aug_op_affine(squeez_img, rotate=x_rot)
        aug_input = unsqueeze_2d_to_3d(rot_squeeze, channels, 'x')
    if y_rot != 0:
        squeez_img = squeeze_3d_to_2d(aug_input, axis_squeeze='y')
        rot_squeeze = aug_op_affine(squeez_img, rotate=y_rot)
        aug_input = unsqueeze_2d_to_3d(rot_squeeze, channels, 'y')

    return aug_input


def aug_cut_out(aug_input: Tensor, fill: float = None, size: int = 16) -> Tensor:
    """
    removing small patch of the image. https://arxiv.org/abs/1708.04552
    :param aug_input: the tensor to augment
    :param fill: value to fill the patch
    :param size:  size of patch
    :return: the augmented tensor
    """
    fill = aug_input.mean(-1).mean(-1) if fill is None else fill
    sx = torch.randint(0, aug_input.shape[1] - size, (1,))
    sy = torch.randint(0, aug_input.shape[2] - size, (1,))
    aug_input[:, sx:sx + size, sy:sy + size] = fill[:, None, None]

    return aug_input


def aug_op_batch_mix_up(aug_input: Tuple[Tensor, Tensor], factor: float) -> Tuple[Tensor, Tensor]:
    """
    mixup augmentation on a batch level
    :param aug_input: batch level input to augment. tuple of image and one hot vector of targets
    :param factor: background factor
    :return: the augmented batch
    """
    img = aug_input[0]
    labels = aug_input[1]
    perm = numpy.arange(img.shape[0])
    numpy.random.shuffle(perm)
    img_mix_up = img[perm]
    labels_mix_up = labels[perm]
    img = img * (1.0 - factor) + factor * img_mix_up
    labels = labels * (1.0 - factor) + factor * labels_mix_up
    return img, labels



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

try:
    from torch.hub import load_state_dict_from_url
except:
    pass

import logging
from typing import Optional
from fuse.models.backbones.backbone_inception_resnet_v2 import FuseBackboneInceptionResnetV2


class InceptionResnetV2(FuseBackboneInceptionResnetV2):
    """
    2D ResNet backbone
    """

    def __init__(self, pretrained_weights_url: Optional[str] = 'http://data.lip6.fr/cadene/pretrainedmodels/inceptionresnetv2-520b38e4.pth',
                 input_channels_num: int = 3) -> None:
        """
        Create InceptionResnetV2
        :param pretrained_weights_url: url containing pretrained weights, if None pretrained weights won't be loaded
        :param input_channels_num:     number of channels in the input data
        """

        # init original model
        super().__init__(logical_units_num=43, input_channels_num=3)

        # load pretrained parameters if required
        if pretrained_weights_url is not None:
            try:
                state_dict = load_state_dict_from_url(pretrained_weights_url)
                self.load_state_dict(state_dict, strict=False)
            except AttributeError:
                logger = logging.getLogger('Fuse')
                logger.info('Invalid URL for InceptionResnetV2 pretrained weights')

        # save input parameters
        self.input_channels_num = input_channels_num




"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Any

import matplotlib.pyplot as plt
import numpy as np

from fuse.data.visualizer.visualizer_base import FuseVisualizerBase
from fuse.utils.utils_hierarchical_dict import FuseUtilsHierarchicalDict


class FuseVisualizerImageAnalysis(FuseVisualizerBase):
    """
    Class for producing analysis of an image
    """

    def __init__(self, image_name: str):
        """
        :param image_name: pointer to an image in batch_dict

        """
        self.image_name = image_name

    def visualize(self, sample: Any, block: bool = True):
        """
                visualize sample
                :param sample: batch_dict - to extract the sample from
                :param block: set to False if the process should not be blocked until the plot will be closed
                :return: None
                """
        # extract data
        image = FuseUtilsHierarchicalDict.get(sample, self.image_name)
        image = image.numpy()
        num_channels = image.shape[0]
        for i in range(num_channels):
            channel_image = image[i, ...]
            if len(channel_image.shape) == 3:
                self.visualize_3dimage(channel_image, title="Image and its Histogram of channel:" + str(i), block=block)
            else:
                assert len(channel_image.shape) == 2
                self.visualise_2dimage(channel_image, title="Image and its Histogram of channel:" + str(i), block=block)

    def visualize_3dimage(self, image: np.array, title: str = "Image and its Histogram", bins=256, block: bool = True) -> None:
        def key_event(e, curr_pos):
            if e.key == "right":
                curr_pos[0] = curr_pos[0] + 1
            elif e.key == "left":
                curr_pos[0] = curr_pos[0] - 1
            else:
                return
            curr_pos[0] = curr_pos[0] % image.shape[0]

            axs[0].cla()
            axs[1].cla()
            axs[0].imshow(image[curr_pos[0]])
            axs[1].hist(image[curr_pos[0]].ravel(), bins=bins, fc='k', ec='k')
            fig.canvas.draw()
            plt.suptitle(title + " at slice:" + str(curr_pos[0]))

        fig, axs = plt.subplots(2)
        position_list = [0]
        fig.canvas.mpl_connect('key_press_event', lambda event: key_event(event, position_list))
        axs[0].imshow(image[0])
        axs[1].hist(image.ravel(), bins=bins, fc='k', ec='k')  # calculating histogram
        plt.suptitle(title + " at slice:" + str(position_list[0]))
        plt.show()

    def visualise_2dimage(self, image: np.array, title: str = "Image and its Histogram", block: bool = True) -> None:
        """
        visualize sample
        :param image: image in the form of np.array
        :param block: set to False if the process should not be blocked until the plot will be closed
        :return: None
        """
        fig = plt.figure()
        fig.add_subplot(221)
        plt.title('image')
        plt.imshow(image)

        fig.add_subplot(222)
        plt.title('histogram')
        plt.hist(image.ravel(), bins=256, fc='k', ec='k')  # calculating histogram

        plt.suptitle(title)
        plt.show(block=block)

    def visualize_aug(self, orig_sample: dict, aug_sample: dict, block: bool = True) -> None:
        """
        Visualise and compare augmented and non-augmented version of the sample
        :param orig_sample: original sample
        :param aug_sample: augmented sample
        :param block: set to False if the process should not be blocked until the plot will be closed
        :return: None
        """
        raise NotImplementedError



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

"""
Augmentor Default class
"""
from typing import Any, Iterable

from fuse.data.augmentor.augmentor_base import FuseAugmentorBase
from fuse.utils.utils_hierarchical_dict import FuseUtilsHierarchicalDict
from fuse.utils.utils_logger import log_object_input_state, convert_state_to_str
from fuse.utils.utils_param_sampler import sample_all


class FuseAugmentorDefault(FuseAugmentorBase):
    """
    Default generic implementation for Fuse augmentor. Aimed to be used by most experiments.
    """

    def __init__(self, augmentation_pipeline: Iterable[Any] = ()):
        """
        :param augmentation_pipeline: list of augmentation operation description,
        Each operation description expected to be a tuple of 4 elements:
            Element 0 - the sample keys affected by this operation
            Element 1 - callback to a function performing the operation. This function expected to support input parameter 'aug_ingput'
            Element 2 - dictionary including the input parameters for the callback function. See AugmentorSamplerDefault
                        to learn how to use random numbers
            Element 3 - general parameters: TBD

            Example:
                See in aug_image_default_pipeline()
        """
        # log object input state
        log_object_input_state(self, locals())

        self.augmentation_pipeline = augmentation_pipeline

    def get_random_augmentation_desc(self) -> Any:
        """
        See description in super class.
        """
        return sample_all(self.augmentation_pipeline)

    def apply_augmentation(self, sample: Any, augmentation_desc: Any) -> Any:
        """
        See description in super class.
        """
        aug_sample = sample
        for op_desc in augmentation_desc:
            # decode augmentation description
            sample_keys = op_desc[0]
            augment_function = op_desc[1]
            augment_function_parameters = op_desc[2]
            general_parameters: dict = op_desc[3]

            # If apply sampled as False skip - by default it will always be True
            apply = general_parameters.get('apply', True)
            if not apply:
                continue

            # Extract augmentation input
            if sample_keys is None:
                aug_input = aug_sample
            elif len(sample_keys) == 1:
                aug_input = FuseUtilsHierarchicalDict.get(aug_sample, sample_keys[0])
            else:
                aug_input = tuple((FuseUtilsHierarchicalDict.get(aug_sample, key) for key in sample_keys))
            augment_function_parameters = augment_function_parameters.copy()
            augment_function_parameters['aug_input'] = aug_input

            # apply augmentation
            aug_result = augment_function(**augment_function_parameters)

            # modify the sample accordingly
            if sample_keys is None:
                aug_sample = aug_result
            elif len(sample_keys) == 1:
                FuseUtilsHierarchicalDict.set(aug_sample, sample_keys[0], aug_result)
            else:
                for index, key in enumerate(sample_keys):
                    FuseUtilsHierarchicalDict.set(aug_sample, key, aug_result[index])

        return aug_sample

    def summary(self) -> str:
        """
        String summary of the object
        """
        return \
            f'Class = {self, __class__}\n' \
                f'Pipeline = {convert_state_to_str(self.augmentation_pipeline)}'



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Sequence, Dict, Tuple

import torch

from fuse.models.model_default import FuseModelDefault
from fuse.models.backbones.backbone_inception_resnet_v2 import FuseBackboneInceptionResnetV2
from fuse.models.heads.head_global_pooling_classifier import FuseHeadGlobalPoolingClassifier
from fuse.utils.utils_hierarchical_dict import FuseUtilsHierarchicalDict


class FuseModelSiamese(FuseModelDefault):
    """
    Fuse Siamese model - 2 branches of the same convolutional neural network with multiple heads
    """

    def __init__(self,
                 conv_inputs_0: Tuple[Tuple[str, int], ...] = (('data.input.input_0.tensor', 1),),
                 conv_inputs_1: Tuple[Tuple[str, int], ...] = (('data.input.input_1.tensor', 1),),
                 backbone: torch.nn.Module = FuseBackboneInceptionResnetV2(),
                 heads: Sequence[torch.nn.Module] = (FuseHeadGlobalPoolingClassifier(),)
                 ) -> None:
        """
        Fuse Siamese model -  two branches with same convolutional neural network with multiple heads
        :param conv_inputs:     batch_dict name for model input and its number of input channels
        :param backbone:        PyTorch backbone module - a convolutional neural network
        :param heads:           Sequence of head modules
            """
        super().__init__((), backbone, heads)
        self.conv_inputs_0 = conv_inputs_0
        self.conv_inputs_1 = conv_inputs_1

    def forward(self,
                batch_dict: Dict) -> Dict:
        backbone_input_0 = [FuseUtilsHierarchicalDict.get(batch_dict, conv_input[0]) for conv_input in self.conv_inputs_0]  #[batch,channels(1),h,w]
        backbone_input_1 = [FuseUtilsHierarchicalDict.get(batch_dict, conv_input[0]) for conv_input in self.conv_inputs_1]  #[batch,channels(1),h,w]

        backbone_features_0 = self.backbone.forward(torch.stack(backbone_input_0, dim=1))  #batch, features, h', w'
        backbone_features_1 = self.backbone.forward(torch.stack(backbone_input_1, dim=1))  #batch, features, h', w'

        backbone_features = torch.cat([backbone_features_0, backbone_features_1], dim=1)

        FuseUtilsHierarchicalDict.set(batch_dict, 'model.backbone_features', backbone_features)
        FuseUtilsHierarchicalDict.set(batch_dict, 'model.backbone_features_0', backbone_features_0)
        FuseUtilsHierarchicalDict.set(batch_dict, 'model.backbone_features_1', backbone_features_1)

        for head in self.heads:
            batch_dict = head.forward(batch_dict)

        return batch_dict['model']



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import logging

import numpy as np
import sklearn.metrics
from typing import Dict, Optional, Callable, Union, List, Tuple

from Fuse.metrics.metric_base import FuseMetricBase
from Fuse.metrics.metrics_toolbox import FuseMetricsToolBox


class FuseMetricConfusionMatrix(FuseMetricBase):
    """
    Computes the confusion matrix from the data
    Multi class version
    Will add to epoch results:
    metrics.<metric_name>.<cls> - array of predicted labels for each class index out of the true labels
        e.g.,
            {metrics.<metric_name>.'class_0': array([2, 2, 1]),
            metrics.<metric_name>.'class_1': array([0, 1, 2]),
            metrics.<metric_name>.'class_2': array([1, 0, 1])}
        here, out of 5 known class_0 labels: 2 are true prediction, 2 were predicted to class_1, 1 to class_1
    """

    def __init__(self, pred_name: str,
                 target_name: str,
                 class_names: Optional[List[str]] = None,
                 operation_point:Optional[Union[float, List[Tuple], Callable]] = None,
                 filter_func: Optional[Callable] = None,
                 use_sample_weights: Optional[bool] = False,
                 sample_weight_name: Optional[str] = None) -> None:
        """
        :param pred_name:       batch_dict key for predicted output (e.g., class probabilities after softmax)
        :param target_name:     batch_dict key for target (e.g., ground truth label)
        :param class_names:     Optional - name for each class otherwise the class index will be used.
        :param operation_point: Optional - few possible options:
                                    (1) None - the predicted class will be selected using argmax
                                    (2) float - binary classification, single threshold for the positive class/
                                    (3) threshold for each class, the order of the list is the order in which class
                                    is assigned, Each element in the list is a tuple of (class_idx, class_threshold)
                                    (4) Callable - getting as an input the prediction and targets and returns the format specified in (3)
        :param filter_func:     Optional - function that filters batch_dict. The function gets as input batch_dict and returns filtered batch_dict.
                                When None (default), the entire batch dict is collected and processed.
        :param use_sample_weights:  Optional - when True (default is False), metrics are computed with sample weigths.
                                    For this, use argument 'sample_weight_name' to define the batch_dict key for the sample weight
        :param sample_weight_name:  Optional, when use_sample_weights is True, user must specify the key for collecting sample weights
        """
        super().__init__(pred_name, target_name, filter_func=filter_func,
                         use_sample_weights=use_sample_weights, sample_weight_name=sample_weight_name)
        self._class_names = class_names
        self._operation_point = operation_point


    def process(self) -> Dict[str, float]:
        """
        Calculates Confusion Matrix per epoch
        :return: dictionary including Area under ROC curve (floating point in range [0, 1]/-1) for each class (class vs rest),
                 -1 will be set for invalid/undefined result (an error message will be printed)
                 The dictionary will also include the average AUC
        """
        # find operation points
        if isinstance(self._operation_point, Callable):
            class_thresholds = self._operation_point(self.epoch_preds, self.epoch_targets)
        elif isinstance(self._operation_point, float):
            class_thresholds = [(1, self._operation_point), (0, 0.0)]
        else:
            class_thresholds = self._operation_point

        results = {}
        # convert probabilities to class predictions
        epoch_class_preds = FuseMetricsToolBox.convert_probabilities_to_class(self.epoch_preds, class_thresholds)

        # set class names if not provided by the user
        class_names = FuseMetricsToolBox.get_class_names(self.epoch_preds[0].shape[-1], self._class_names)

        sample_weights = self.collected_data['sample_weight'] if self.use_sample_weights else None

        conf_matrix = sklearn.metrics.confusion_matrix(self.collected_data['target'], epoch_class_preds, sample_weight=sample_weights)
        for cls_ind, cls in enumerate(class_names):
            results[cls] = conf_matrix[cls_ind]

        return results


if __name__ == '__main__':
    data = {'preds': np.array([[0.8, 0.1, 0.1],
                               [0.5, 0.3, 0.2],
                               [0.6, 0.3, 0.1],
                               [0.6, 0.1, 0.3],
                               [0.7, 0.2, 0.1],
                               [0.3, 0.2, 0.5],
                               [0.1, 0.2, 0.7],
                               [0.2, 0.6, 0.2],
                               [0.3, 0.3, 0.4],
                               [0.7, 0.2, 0.1]]),
            'targets': np.array([0, 0, 0, 0, 0, 1, 1, 1, 2, 2]),
            'weights': np.array([0.03, 0.9, 0.05, 0.52, 0.23, 0.72, 0.13, 0.113, 0.84, 0.09])}

    print("threshold=((0, 0.6), (1, 0.5), (2, 0.2))")
    metric = FuseMetricConfusionMatrix(pred_name='preds', target_name='targets',
                                       operation_point=[(0, 0.6), (1, 0.5), (2, 0.2)], use_sample_weights=True,
                                       sample_weight_name='weights')
    metric.collect(data)
    res = metric.process()
    print('with weights')
    for k, v in res.items():
        print(k, '\t', v)
    metric = FuseMetricConfusionMatrix(pred_name='preds', target_name='targets',
                                       operation_point=[(0, 0.6), (1, 0.5), (2, 0.2)], use_sample_weights=False,
                                       sample_weight_name='weights')
    metric.collect(data)
    res = metric.process()
    print('no weights')
    for k, v in res.items():
        print(k, '\t', v)

    print('no threshold (argmax)')
    metric = FuseMetricConfusionMatrix(pred_name='preds', target_name='targets')
    metric.collect(data)
    res = metric.process()
    for k, v in res.items():
        print(k, '\t', v)

    data = {'preds': np.array([[0.8, 0.2],
                               [0.5, 0.5],
                               [0.6, 0.4],
                               [0.6, 0.4],
                               [0.7, 0.3],
                               [0.3, 0.7],
                               [0.1, 0.9],
                               [0.2, 0.8],
                               [0.3, 0.7],
                               [0.4, 0.6]]),
            'targets': np.array([0, 0, 1, 0, 0, 1, 0, 1, 1, 0])}
    metric = FuseMetricConfusionMatrix(pred_name='preds', target_name='targets')
    metric.collect(data)
    res = metric.process()
    print('binary')
    for k, v in res.items():
        print(k, '\t', v)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from fuse.metrics.metrics_toolbox import FuseMetricsToolBox

from fuse.metrics.metric_base import FuseMetricBase
from typing import Dict, List, Callable, Optional, Sequence
import numpy as np


class FuseMetricConfusionIndeterminate(FuseMetricBase):
    """
    Multi class version for confusion metrics such as sensitivity,specificity, etc... Given an option to filter out indeterminate predictions.

    """

    def __init__(self, pred_name: str, target_name: str, filter_func: Optional[Callable] = None, class_names: List = None,
                 indeterminate_thresholds: Sequence[float] = (0.5,), indeterminate_classes: Sequence[int] = tuple(),
                 determinate_percent_mode: bool = False,
                 metrics: Sequence[str] = ('sensitivity',), sum_weights: List = None, **kwargs):
        """
        :param pred_name:       batch_dict key for predicted output (e.g., class probabilities after softmax)
        :param target_name:     batch_dict key for target (e.g., ground truth label)
        :param class_names:     Optional - name for each class otherwise the class index will be used.
        :param filter_func: function that filters batch_dict The function gets ans input batch_dict and returns filtered batch_dict
        :param indeterminate_thresholds: threshold for indeterminate class or if determinate_percent_mode set to True: target percent of samples to predict
                                        (the rest will be marked as indeterminate)
        :param indeterminate_classes: if the sample prediction included in indeterminate_classes, it will be marked as indeterminate.
        :param determinate_percent_mode: see indeterminate_thresholds
        :param metrics: list of metrics name to return. See FuseMetricsToolBox.confusion_metrics for available options.
        :param sum_weights: weights for the weighted sum of sensitivity average

        """
        super().__init__(pred_name, target_name, filter_func, **kwargs)
        self._class_names = class_names
        self._indeterminate_thresholds = indeterminate_thresholds
        self._indeterminate_classes = indeterminate_classes
        self._determinate_percent_mode = determinate_percent_mode
        self._metrics = metrics
        self._sum_weights = sum_weights

    def process(self) -> Dict:
        """
        Calculates sensitivity
        :return: dictionary including sensitivity for each class (class vs rest) given indeterminate threshold,
        Expected output per indeterminate threshold:
        '
        ind_th_<indeterminate threshold>.determinate_percentage: <value>
        ind_th_<indeterminate threshold>.<metric>_<class name 0>: <value>
        ind_th_<indeterminate threshold>.<metric>_<class name 1>_sens: <value>
        ...
        ind_th_<indeterminate threshold>.<metric>_macro_avg: <value>
        '

        """
        results = {}

        # set class names if not provided by the user
        class_names = FuseMetricsToolBox.get_class_names(self.epoch_preds[0].shape[-1], self._class_names)

        # set weight per class if not provided by the user
        if self._sum_weights is None:
            self._sum_weights = [1] * len(class_names)
        self._sum_weights = np.array(self._sum_weights)

        epoch_preds = np.array(self.epoch_preds)
        epoch_targets = np.array(self.epoch_targets)
        epoch_class_preds = epoch_preds.argmax(axis=1)
        epoch_score_preds = epoch_preds.max(axis=1)
        epoch_score_preds_lst = [epoch_score_preds[epoch_class_preds == i] for i in range(len(self._class_names))]

        for threshold in self._indeterminate_thresholds:

            # convert percentile to score threshold
            if self._determinate_percent_mode:
                if isinstance(threshold, list):
                    threshold = [np.percentile(preds, (1.0 - threshold[i]) * 100.0) for i, preds in enumerate(epoch_score_preds_lst)]
                else:
                    threshold = np.percentile(epoch_score_preds, (1.0 - threshold) * 100.0)

            # mark prediction lower than the threshold as indeterminate
            if isinstance(threshold, list):
                threshold = np.array(threshold)
                determinate_indices = (epoch_score_preds >= threshold[epoch_class_preds])
            else:
                determinate_indices = epoch_score_preds >= threshold

            # mark indeterminate classes as in indeterminate
            for cls_ind in self._indeterminate_classes:
                determinate_indices = determinate_indices & (epoch_class_preds != cls_ind)

            # filter indeterminate samples
            results[f'ind_th_{threshold}_determinate_percentage'] = np.round(determinate_indices.sum() / epoch_preds.shape[0], 3)
            preds = epoch_class_preds[determinate_indices]
            targets = epoch_targets[determinate_indices]

            if preds.shape[0] == 0:
                for metric_name in self._metrics:
                    results[f'ind_th_{threshold}_{metric_name}_macro_avg'] = 'N/A'

            # compute the confusion metrics
            agg_results = {}
            for cls_ind, cls in enumerate(class_names):
                cls_result = FuseMetricsToolBox.confusion_metrics(preds, cls_ind, targets, self._metrics)

                results[f'ind_th_{threshold}_determinate_percentage_{cls}'] = np.round(
                    np.where(targets == cls_ind, 1, 0).sum() / np.where(epoch_targets == cls_ind, 1, 0).sum(), 3)

                for metric_name in self._metrics:
                    results[f'ind_th_{threshold}_{metric_name}_{cls}'] = cls_result[metric_name]

                # aggregate results
                for metric_name in self._metrics:
                    if metric_name not in agg_results:
                        agg_results[metric_name] = {}
                    metric_result = agg_results[metric_name]
                    metric_result[cls] = cls_result[metric_name]

            # calculate macro average
            for metric_name in self._metrics:
                value_per_class = np.array(list(agg_results[metric_name].values()))
                indices = ~np.isnan(value_per_class)
                results[f'ind_th_{threshold}_{metric_name}_macro_avg'] = np.average(value_per_class[indices], weights=self._sum_weights[indices])

        return results



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

"""

(C) Copyright 2021 IBM Corp.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
   http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Created on June 30, 2021

"""

import os

from fuse.analyzer.analyzer_default import FuseAnalyzerDefault
from fuse.utils.utils_debug import FuseUtilsDebug

from fuse.utils.utils_gpu import FuseUtilsGPU

os.environ['skip_broker'] = '1'

import logging

import torch.optim as optim
import torch.nn.functional as F
from torch.utils.data.dataloader import DataLoader

from fuse.utils.utils_param_sampler import FuseUtilsParamSamplerUniform as Uniform
from fuse.utils.utils_param_sampler import FuseUtilsParamSamplerRandInt as RandInt
from fuse.utils.utils_param_sampler import FuseUtilsParamSamplerRandBool as RandBool
from fuse.utils.utils_logger import fuse_logger_start

from fuse.data.sampler.sampler_balanced_batch import FuseSamplerBalancedBatch
from fuse.data.visualizer.visualizer_default import FuseVisualizerDefault
from fuse.data.augmentor.augmentor_default import FuseAugmentorDefault
from fuse.data.augmentor.augmentor_toolbox import aug_op_affine, aug_op_color, aug_op_gaussian
from fuse.data.dataset.dataset_default import FuseDatasetDefault

from fuse.models.model_default import FuseModelDefault
from fuse.models.backbones.backbone_resnet import FuseBackboneResnet
from fuse.models.heads.head_global_pooling_classifier import FuseHeadGlobalPoolingClassifier

from fuse.losses.loss_default import FuseLossDefault

from fuse.metrics.classification.metric_auc import FuseMetricAUC
from fuse.metrics.classification.metric_accuracy import FuseMetricAccuracy
from fuse.metrics.classification.metric_roc_curve import FuseMetricROCCurve

from fuse.managers.callbacks.callback_tensorboard import FuseTensorboardCallback
from fuse.managers.callbacks.callback_metric_statistics import FuseMetricStatisticsCallback
from fuse.managers.callbacks.callback_time_statistics import FuseTimeStatisticsCallback
from fuse.managers.manager_default import FuseManagerDefault


from fuse_examples.classification.skin_lesion.data_source import FuseSkinDataSource
from fuse_examples.classification.skin_lesion.input_processor import FuseSkinInputProcessor
from fuse_examples.classification.skin_lesion.ground_truth_processor import FuseSkinGroundTruthProcessor
from fuse_examples.classification.skin_lesion.inception import InceptionResnetV2
from fuse_examples.classification.skin_lesion.download import download_and_extract_isic


##########################################
# Debug modes
##########################################
mode = 'default'  # Options: 'default', 'fast', 'debug', 'verbose', 'user'. See details in FuseUtilsDebug
debug = FuseUtilsDebug(mode)

##########################################
# Output Paths
##########################################

DATA_YEAR = '2017'
# TODO: Path to save model
ROOT = 'examples/skin/'
# TODO: Path to store the data
ROOT_DATA = 'examples/skin/data'
# TODO: Name of the experiment
EXPERIMENT = 'InceptionResnetV2_2017_test'
# TODO: Path to cache data
CACHE_PATH = 'examples/skin/'
# TODO: Name of the cached data folder
EXPERIMENT_CACHE = 'ISIC_'+ DATA_YEAR

PATHS = {'data_dir': ROOT_DATA,
         'model_dir': os.path.join(ROOT, EXPERIMENT, 'model_dir'),
         'force_reset_model_dir': True,  # If True will reset model dir automatically - otherwise will prompt 'are you sure' message.
         'cache_dir': os.path.join(CACHE_PATH, EXPERIMENT_CACHE+'_cache_dir'),
         'inference_dir': os.path.join(ROOT, EXPERIMENT, 'infer_dir')}


##########################################
# Train Common Params
##########################################
# ============
# Data
# ============
TRAIN_COMMON_PARAMS = {}
TRAIN_COMMON_PARAMS['data.year'] = DATA_YEAR  # year of the ISIC dataset

TRAIN_COMMON_PARAMS['data.train_num_workers'] = 8
TRAIN_COMMON_PARAMS['data.validation_num_workers'] = 8

TRAIN_COMMON_PARAMS['data.augmentation_pipeline'] = [
    [
        ('data.input.input_0',),
        aug_op_affine,
        {'rotate': Uniform(-180.0, 180.0), 'translate': (RandInt(-50, 50), RandInt(-50, 50)),
         'flip': (RandBool(0.3), RandBool(0.3)), 'scale': Uniform(0.9, 1.1)},
        {'apply': RandBool(0.9)}
    ],
    [
        ('data.input.input_0',),
        aug_op_color,
        {'add': Uniform(-0.06, 0.06), 'mul': Uniform(0.95, 1.05), 'gamma': Uniform(0.9, 1.1),
         'contrast': Uniform(0.85, 1.15)},
        {'apply': RandBool(0.7)}
    ],
    [
        ('data.input.input_0',),
        aug_op_gaussian,
        {'std': 0.03},
        {'apply': RandBool(0.7)}
    ],
]

assert TRAIN_COMMON_PARAMS['data.year'] == '2016' or TRAIN_COMMON_PARAMS['data.year'] == '2017'

# ===============
# Manager - Train
# ===============
num_gpus = 1
TRAIN_COMMON_PARAMS['data.batch_size'] = 8 * num_gpus
TRAIN_COMMON_PARAMS['manager.train_params'] = {
    'num_gpus': num_gpus,
    'num_epochs': 15,
    'virtual_batch_size': 1,  # number of batches in one virtual batch
    'start_saving_epochs': 10,  # first epoch to start saving checkpoints from
    'gap_between_saving_epochs': 100,  # number of epochs between saved checkpoint
}

# best_epoch_source
# if an epoch values are the best so far, the epoch is saved as a checkpoint.
TRAIN_COMMON_PARAMS['manager.best_epoch_source'] = {
    'source': 'metrics.auc.macro_avg',  # can be any key from losses or metrics dictionaries
    'optimization': 'max',  # can be either min/max
    'on_equal_values': 'better',
    # can be either better/worse - whether to consider best epoch when values are equal
}
TRAIN_COMMON_PARAMS['manager.learning_rate'] = 1e-5
TRAIN_COMMON_PARAMS['manager.weight_decay'] = 0.001


#################################
# Train Template
#################################
def run_train(paths: dict, train_common_params: dict):
    # ==============================================================================
    # Logger
    # ==============================================================================
    fuse_logger_start(output_path=paths['model_dir'], console_verbose_level=logging.INFO)
    lgr = logging.getLogger('Fuse')

    # Download data
    download_and_extract_isic(paths['data_dir'], train_common_params['data.year'])

    lgr.info('\nFuse Train', {'attrs': ['bold', 'underline']})

    lgr.info(f'model_dir={paths["model_dir"]}', {'color': 'magenta'})
    lgr.info(f'cache_dir={paths["cache_dir"]}', {'color': 'magenta'})

    #### Train Data

    lgr.info(f'Train Data:', {'attrs': 'bold'})

    ## Create data source:

    input_source_gt = {'2016': os.path.join(paths['data_dir'], 'data/ISIC2016_Training_GroundTruth.csv'),
                       '2017': os.path.join(paths['data_dir'], 'data/ISIC2017_Training_GroundTruth.csv')}
    training_data = {'2016': os.path.join(paths['data_dir'], 'data/ISIC2016_Training_Data/'),
                     '2017': os.path.join(paths['data_dir'], 'data/ISIC2017_Training_Data/')}
    input_source_gt_val = {'2016': input_source_gt['2016'],
                           '2017': os.path.join(paths['data_dir'], 'data/ISIC2017_Validation_GroundTruth.csv')}
    validation_data = {'2016': training_data['2016'],
                       '2017': os.path.join(paths['data_dir'], 'data/ISIC2017_Validation_Data/')}

    partition_file = {'2016': "train_val_split.pickle",
                      '2017': None}  # as no validation set is available in 2016 dataset, it needs to be created

    train_data_source = FuseSkinDataSource(input_source_gt[train_common_params['data.year']],
                                           partition_file=partition_file[train_common_params['data.year']],
                                           train=True,
                                           override_partition=True)

    ## Create data processors:
    input_processors = {
        'input_0': FuseSkinInputProcessor(input_data=training_data[train_common_params['data.year']])
    }
    gt_processors = {
        'gt_global': FuseSkinGroundTruthProcessor(input_data=input_source_gt[train_common_params['data.year']],
                                                  year=train_common_params['data.year'])
    }

    # Create data augmentation (optional)
    augmentor = FuseAugmentorDefault(
        augmentation_pipeline=train_common_params['data.augmentation_pipeline'])

    # Create visualizer (optional)
    visualiser = FuseVisualizerDefault(image_name='data.input.input_0', label_name='data.gt.gt_global.tensor')

    # Create dataset
    train_dataset = FuseDatasetDefault(cache_dest=paths['cache_dir'],
                                       data_source=train_data_source,
                                       input_processors=input_processors,
                                       gt_processors=gt_processors,
                                       augmentor=augmentor,
                                       visualizer=visualiser)

    lgr.info(f'- Load and cache data:')
    train_dataset.create()
    lgr.info(f'- Load and cache data: Done')

    ## Create sampler
    lgr.info(f'- Create sampler:')
    sampler = FuseSamplerBalancedBatch(dataset=train_dataset,
                                       balanced_class_name='data.gt.gt_global.tensor',
                                       num_balanced_classes=2,
                                       batch_size=train_common_params['data.batch_size'])

    lgr.info(f'- Create sampler: Done')

    ## Create dataloader
    train_dataloader = DataLoader(dataset=train_dataset,
                                  shuffle=False, drop_last=False,
                                  batch_sampler=sampler, collate_fn=train_dataset.collate_fn,
                                  num_workers=train_common_params['data.train_num_workers'])
    lgr.info(f'Train Data: Done', {'attrs': 'bold'})

    #### Validation data
    lgr.info(f'Validation Data:', {'attrs': 'bold'})

    ## Create data source
    validation_data_source = FuseSkinDataSource(input_source_gt_val[train_common_params['data.year']],
                                                partition_file=partition_file[train_common_params['data.year']],
                                                train=False)

    input_processors = {
        'input_0': FuseSkinInputProcessor(input_data=validation_data[train_common_params['data.year']])
    }
    gt_processors = {
        'gt_global': FuseSkinGroundTruthProcessor(input_data=input_source_gt_val[train_common_params['data.year']],
                                                  year=train_common_params['data.year'])
    }

    ## Create dataset
    validation_dataset = FuseDatasetDefault(cache_dest=paths['cache_dir'],
                                            data_source=validation_data_source,
                                            input_processors=input_processors,
                                            gt_processors=gt_processors,
                                            augmentor=None,
                                            visualizer=visualiser)

    lgr.info(f'- Load and cache data:')
    validation_dataset.create(pool_type='thread')  # use ThreadPool to create this dataset, to avoid cv2 problems in multithreading
    lgr.info(f'- Load and cache data: Done')

    ## Create dataloader
    validation_dataloader = DataLoader(dataset=validation_dataset,
                                       shuffle=False,
                                       drop_last=False,
                                       batch_sampler=None,
                                       batch_size=train_common_params['data.batch_size'],
                                       num_workers=train_common_params['data.validation_num_workers'],
                                       collate_fn=validation_dataset.collate_fn)
    lgr.info(f'Validation Data: Done', {'attrs': 'bold'})

    # ==============================================================================
    # Model
    # ==============================================================================
    lgr.info('Model:', {'attrs': 'bold'})

    model = FuseModelDefault(
        conv_inputs=(('data.input.input_0', 1),),
        backbone={'Resnet18': FuseBackboneResnet(pretrained=True, in_channels=3, name='resnet18'),
                  'InceptionResnetV2': InceptionResnetV2()}['InceptionResnetV2'],
        heads=[
            FuseHeadGlobalPoolingClassifier(head_name='head_0',
                                            dropout_rate=0.5,
                                            conv_inputs=[('model.backbone_features', 1536)],
                                            post_concat_inputs=None,
                                            num_classes=2,
                                            pooling="avg"),
        ]
    )
    lgr.info('Model: Done', {'attrs': 'bold'})


    # ====================================================================================
    #  Loss
    # ====================================================================================
    losses = {
        'cls_loss': FuseLossDefault(pred_name='model.logits.head_0', target_name='data.gt.gt_global.tensor',
                                    callable=F.cross_entropy, weight=1.0)
    }

    # ====================================================================================
    # Metrics
    # ====================================================================================
    metrics = {
        'auc': FuseMetricAUC(pred_name='model.output.head_0', target_name='data.gt.gt_global.tensor'),
        'accuracy': FuseMetricAccuracy(pred_name='model.output.head_0', target_name='data.gt.gt_global.tensor')
    }

    # =====================================================================================
    #  Callbacks
    # =====================================================================================
    callbacks = [
        # default callbacks
        FuseTensorboardCallback(model_dir=paths['model_dir']),  # save statistics for tensorboard
        FuseMetricStatisticsCallback(output_path=paths['model_dir'] + "/metrics.csv"),  # save statistics in a csv file
        FuseTimeStatisticsCallback(num_epochs=train_common_params['manager.train_params']['num_epochs'], load_expected_part=0.1)  # time profiler
    ]

    # =====================================================================================
    #  Manager - Train
    #  Create a manager, training objects and run a training process.
    # =====================================================================================
    lgr.info('Train:', {'attrs': 'bold'})

    # create optimizer
    optimizer = optim.Adam(model.parameters(), lr=train_common_params['manager.learning_rate'],
                           weight_decay=train_common_params['manager.weight_decay'])

    # create scheduler
    scheduler = {'ReduceLROnPlateau': optim.lr_scheduler.ReduceLROnPlateau(optimizer),
                 'CosineAnnealing': optim.lr_scheduler.CosineAnnealingWarmRestarts(optimizer, T_0=1)}['ReduceLROnPlateau']

    # train from scratch
    manager = FuseManagerDefault(output_model_dir=paths['model_dir'], force_reset=paths['force_reset_model_dir'])
    # Providing the objects required for the training process.
    manager.set_objects(net=model,
                        optimizer=optimizer,
                        losses=losses,
                        metrics=metrics,
                        best_epoch_source=train_common_params['manager.best_epoch_source'],
                        lr_scheduler=scheduler,
                        callbacks=callbacks,
                        train_params=train_common_params['manager.train_params'],
                        output_model_dir=paths['model_dir'])

    # Start training
    manager.train(train_dataloader=train_dataloader,
                  validation_dataloader=validation_dataloader)

    lgr.info('Train: Done', {'attrs': 'bold'})



######################################
# Inference Common Params
######################################
INFER_COMMON_PARAMS = {}
INFER_COMMON_PARAMS['infer_filename'] = 'validation_set_infer.gz'
INFER_COMMON_PARAMS['checkpoint'] = 'best'  # Fuse TIP: possible values are 'best', 'last' or epoch_index.
INFER_COMMON_PARAMS['data.year'] = TRAIN_COMMON_PARAMS['data.year']
INFER_COMMON_PARAMS['data.train_num_workers'] = TRAIN_COMMON_PARAMS['data.train_num_workers']


######################################
# Inference Template
######################################
def run_infer(paths: dict, infer_common_params: dict):
    #### Logger
    fuse_logger_start(output_path=paths['inference_dir'], console_verbose_level=logging.INFO)
    lgr = logging.getLogger('Fuse')
    lgr.info('Fuse Inference', {'attrs': ['bold', 'underline']})
    lgr.info(f'infer_filename={os.path.join(paths["inference_dir"], infer_common_params["infer_filename"])}', {'color': 'magenta'})

    # Create data source:
    input_source = {'2016': os.path.join(paths['data_dir'], 'data/ISIC2016_Test_GroundTruth.csv'),
                    '2017': os.path.join(paths['data_dir'], 'data/ISIC2017_Test_GroundTruth.csv')}
    infer_data = {'2016': os.path.join(paths['data_dir'], 'data/ISIC2016_Test_Data/'),
                  '2017': os.path.join(paths['data_dir'], 'data/ISIC2017_Test_Data/')}

    infer_data_source = FuseSkinDataSource(input_source[infer_common_params['data.year']])

    # Create data processors
    input_processors_infer = {
        'input_0': FuseSkinInputProcessor(input_data=infer_data[infer_common_params['data.year']])
    }
    gt_processors_infer = {
        'gt_global': FuseSkinGroundTruthProcessor(input_data=input_source[infer_common_params['data.year']],
                                                  train=False, year=infer_common_params['data.year'])
    }

    # Create visualizer (optional)
    visualiser = FuseVisualizerDefault(image_name='data.input.input_0', label_name='data.gt.gt_global.tensor')

    # Create dataset
    infer_dataset = FuseDatasetDefault(cache_dest=paths['cache_dir'],
                                       data_source=infer_data_source,
                                       input_processors=input_processors_infer,
                                       gt_processors=gt_processors_infer,
                                       visualizer=visualiser)
    infer_dataset.create()

    lgr.info(f'- Create sampler: Done')

    ## Create dataloader
    infer_dataloader = DataLoader(dataset=infer_dataset,
                                  shuffle=False, drop_last=False,
                                  collate_fn=infer_dataset.collate_fn,
                                  num_workers=infer_common_params['data.train_num_workers'])
    lgr.info(f'Test Data: Done', {'attrs': 'bold'})

    #### Manager for inference
    manager = FuseManagerDefault()
    # extract just the global classification per sample and save to a file
    output_columns = ['model.output.head_0']
    manager.infer(data_loader=infer_dataloader,
                  input_model_dir=paths['model_dir'],
                  checkpoint=infer_common_params['checkpoint'],
                  output_columns=output_columns,
                  output_file_name=os.path.join(paths["inference_dir"], infer_common_params["infer_filename"]))


######################################
# Analyze Common Params
######################################
ANALYZE_COMMON_PARAMS = {}
ANALYZE_COMMON_PARAMS['infer_filename'] = INFER_COMMON_PARAMS['infer_filename']
ANALYZE_COMMON_PARAMS['output_filename'] = 'all_metrics.txt'
ANALYZE_COMMON_PARAMS['num_workers'] = 4
ANALYZE_COMMON_PARAMS['batch_size'] = 8
ANALYZE_COMMON_PARAMS['data.year'] = TRAIN_COMMON_PARAMS['data.year']


######################################
# Analyze Template
######################################
def run_analyze(paths: dict, analyze_common_params: dict):
    fuse_logger_start(output_path=None, console_verbose_level=logging.INFO)
    lgr = logging.getLogger('Fuse')
    lgr.info('Fuse Analyze', {'attrs': ['bold', 'underline']})

    # dataset generating ground truth processors
    input_source = {'2016': os.path.join(paths['data_dir'], 'data/ISIC2016_Test_GroundTruth.csv'),
                    '2017': os.path.join(paths['data_dir'], 'data/ISIC2017_Test_GroundTruth.csv')}

    ## Create data processors
    gt_processors = {
        'gt_global': FuseSkinGroundTruthProcessor(input_data=input_source[analyze_common_params['data.year']],
                                                  train=False, year=analyze_common_params['data.year'])
    }

    # metrics
    metrics = {
        'auc': FuseMetricAUC(pred_name='model.output.head_0', target_name='data.gt.gt_global.tensor'),
        'accuracy': FuseMetricAccuracy(pred_name='model.output.head_0', target_name='data.gt.gt_global.tensor'),
        'roc': FuseMetricROCCurve(pred_name='model.output.head_0', target_name='data.gt.gt_global.tensor',
                                  output_filename=os.path.join(paths["inference_dir"], "roc_curve.png"))
    }

    # create analyzer
    analyzer = FuseAnalyzerDefault()

    # run
    results = analyzer.analyze(gt_processors=gt_processors,
                               data_pickle_filename=os.path.join(paths["inference_dir"], analyze_common_params["infer_filename"]),
                               metrics=metrics,
                               output_filename=os.path.join(paths["inference_dir"], analyze_common_params['output_filename']),
                               num_workers=analyze_common_params['num_workers'],
                               batch_size=analyze_common_params['num_workers'])

    return results


######################################
# Run
######################################


if __name__ == "__main__":
    # allocate gpus
    NUM_GPUS = 1
    if NUM_GPUS == 0:
        TRAIN_COMMON_PARAMS['manager.train_params']['device'] = 'cpu'
    # uncomment if you want to use specific gpus instead of automatically looking for free ones
    force_gpus = None  # [0]
    FuseUtilsGPU.choose_and_enable_multiple_gpus(NUM_GPUS, force_gpus=force_gpus)

    RUNNING_MODES = ['train', 'infer', 'analyze']  # Options: 'train', 'infer', 'analyze'

    # train
    if 'train' in RUNNING_MODES:
        run_train(paths=PATHS, train_common_params=TRAIN_COMMON_PARAMS)

    # infer
    if 'infer' in RUNNING_MODES:
        run_infer(paths=PATHS, infer_common_params=INFER_COMMON_PARAMS)

    # analyze
    if 'analyze' in RUNNING_MODES:
        run_analyze(paths=PATHS, analyze_common_params=ANALYZE_COMMON_PARAMS)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import logging
import os
import traceback

import numpy as np
import pandas as pd
import torch
import torch.nn as nn
from torch.optim.optimizer import Optimizer
from torch.utils.data.dataloader import DataLoader
from tqdm import trange, tqdm
from typing import Dict, Any, List, Iterator, Optional, Union, Sequence, Hashable, Callable

from fuse.data.data_source.data_source_base import FuseDataSourceBase
from fuse.data.dataset.dataset_base import FuseDatasetBase
from fuse.data.processor.processor_base import FuseProcessorBase
from fuse.data.visualizer.visualizer_base import FuseVisualizerBase
from fuse.losses.loss_base import FuseLossBase
from fuse.managers.callbacks.callback_base import FuseCallback
from fuse.managers.callbacks.callback_debug import FuseCallbackDebug
from fuse.managers.callbacks.callback_infer_results import FuseInferResultsCallback
from fuse.managers.manager_state import FuseManagerState
from fuse.metrics.metric_base import FuseMetricBase
from fuse.models.model_ensemble import FuseModelEnsemble
from fuse.utils import utils_misc as misc
from fuse.utils.utils_checkpoint import FuseCheckpoint
from fuse.utils.utils_debug import FuseUtilsDebug
from fuse.utils.utils_file import FuseUtilsFile as file
from fuse.utils.utils_gpu import FuseUtilsGPU as gpu
from fuse.utils.utils_hierarchical_dict import FuseUtilsHierarchicalDict
from fuse.utils.utils_logger import log_object_input_state
from fuse.utils.utils_misc import FuseUtilsMisc


class FuseManagerDefault:
    """
    Default implementation of manager. The manager is the main API to use when running Fuse.
    Supports Train and Infer functionality.


    Possible Work flows for using the manager are (see function documentations for parameters description):
    For train:
        FuseManagerDefault() -> manager.set_objects() -> manager.train()
    For Resume training:
        FuseManagerDefault() -> manager.load_objects() -> manager.load_checkpoint() -> manager.train()
    For Train using existing model:
        FuseManagerDefault() -> manager.set_objects() [-> manager.load_objects()] [-> manager.load_checkpoint()] -> manager.train()
    For Infer:
        FuseManagerDefault() -> manager.infer()
        or -
        FuseManagerDefault() -> manager.load_objects() -> manager.load_checkpoint() -> manager.infer()
    For Infer given model:
        FuseManagerDefault() -> manager.set_objects() -> manager.load_checkpoint() -> manager.infer()
    """

    def __init__(self, output_model_dir: str = None, force_reset: bool = False):
        """
        Initialization of the manager.
        Responsible for creating the class members, creating the output_model_dir if not None.

        :param output_model_dir: path to the model directory
        :param force_reset: When False (default) and model_dir content has to be reset the user is prompt to do so.
            When True, the directory is automatically reset.
        """
        log_object_input_state(self, locals())
        self.logger = logging.getLogger('Fuse')

        self.state = FuseManagerState()
        self.state.output_model_dir = output_model_dir
        self.state.current_epoch = 0

        if output_model_dir is not None:
            # prepare model_dir
            file.create_or_reset_dir(output_model_dir, ignore_files=['logs', 'source_files'], force_reset=force_reset)

        self.callbacks: List[FuseCallback] = list()  # callback can be empty
        pass

    def set_objects(self,
                    net: nn.Module = None,
                    # ensemble_nets: Sequence[nn.Module] = None,
                    metrics: Dict[str, FuseMetricBase] = None,
                    losses: Dict[str, FuseLossBase] = None,
                    callbacks: List[FuseCallback] = None,
                    optimizer: Optimizer = None,
                    lr_scheduler: Any = None,
                    best_epoch_source: Union[List[Dict[str, str]], Dict[str, str]] = None,
                    train_params: Dict = None,
                    output_model_dir: str = None) -> None:
        """
        Sets objects to the given values.
        :param net: the model definition
        :param metrics: a dictionary of metric name and FuseMetricBase to compute on train and validation sets.
        :param losses: definition of loss functions
        :param callbacks: list of callbacks to call during train or infer.
        :param optimizer: optimizer to use
        :param lr_scheduler: learning rate scheduler to use
        :param best_epoch_source: metric or loss function to use when deciding on the best epoch. may also be a list of them
                        contain the keys:
                         'source': name of loss or metric function - e.g. losses.cls_loss or metrics.auc
                         'optimization': the optimization function used on the source function. either min or max.
                         'on_equal_values': in case where the current value is equal the best so far value should it be considered as best epoch?
                                            value can be either better or worse.
        :param train_params: dictionary containing configuration data (mostly for training).
            may contain the keys (otherwise filled with defaults):
            device - options: 'cuda', 'cpu', 'cuda:0', ... (default 'cuda')
            num_epochs - number of epochs to run (default 100),
            virtual_batch_size - number of batches in one virtual batch (default 1),
            start_saving_epochs - first epoch to start saving checkpoint (default 80)
            gap_between_saving_epochs - number of epochs between each saved checkpoint
        :param output_model_dir: directory to save the model data to

        """
        best_epoch_source = [best_epoch_source] if not isinstance(best_epoch_source, list) else best_epoch_source

        if net is not None: self.state.net = net
        # if ensemble_nets is not None: self.state.ensemble_nets = ensemble_nets
        if metrics is not None: self.state.metrics = metrics
        if losses is not None: self.state.losses = losses
        if callbacks is not None: self.callbacks = callbacks
        if optimizer is not None: self.state.optimizer = optimizer
        if lr_scheduler is not None: self.state.lr_scheduler = lr_scheduler
        if best_epoch_source is not None: self.state.best_epoch_source = best_epoch_source
        if train_params is not None: self.state.train_params = train_params
        if output_model_dir is not None: self.state.output_model_dir = output_model_dir

        # debug mode - append debug callback
        if FuseUtilsDebug().get_setting('manager_stages') != 'default':
            self.callbacks.append(FuseCallbackDebug())
            self.logger.info(f'Manager - debug mode - append debug callback', {'color': 'red'})
        pass

    def _save_objects(self, validation_dataloader: DataLoader) -> None:
        """
        Saves objects using torch.save (net, losses, metrics, best_epoch_source, optimizer, lr_scheduler, callbacks).
        Each parameter is saved into a separate file (called losses.pth, metrics.pth, etc) under self.output_model_dir.
        :param validation_dataloader: dataloader to extract dataset definitions from (saved on inference_dataset.pth)
        """

        def _torch_save(parameter_to_save: Any, parameter_name: str) -> None:
            file_path = os.path.join(self.state.output_model_dir, f'{parameter_name}.pth')
            torch.save(parameter_to_save, file_path)
            self.logger.debug(f"Saved object to file {file_path}", {'color': 'green'})
            pass

        _torch_save(self.state.net.module, 'net')
        _torch_save(self.state.metrics, 'metrics')
        _torch_save(self.state.losses, 'losses')
        _torch_save(self.callbacks, 'callbacks')
        _torch_save(self.state.optimizer, 'optimizer')
        _torch_save(self.state.lr_scheduler, 'lr_scheduler')
        _torch_save(self.state.best_epoch_source, 'best_epoch_source')
        _torch_save(self.state.train_params, 'train_params')

        # also save validation_dataset in inference mode
        if validation_dataloader is not None:
            FuseDatasetBase.save(validation_dataloader.dataset, mode=FuseDatasetBase.SaveMode.INFERENCE,
                                 filename=os.path.join(self.state.output_model_dir, "inference_dataset.pth"))
        pass

    def load_objects(self, input_model_dir: Union[str, Sequence[str]], list_of_object_names: List[str] = None, mode: str = 'infer') -> Dict[str, Any]:
        """
        Loads objects from torch saved pth files under input_model_dir.
        Note: if mode is infer, only loads net.
        :param input_model_dir: the path to the files to load. if list/tuple: load multiple modules for ensemble inference
        :param list_of_object_names: list of the objects to load. If None, then all objects are loaded.
            For the possible values of objects to be loaded, see input parameters of setObjects().
        :param mode: either 'infer' or 'train'. When 'infer' (default), only net is loaded.
        :return: for each loaded object there's a key and the value is the member variable
        """

        def should_load(object_name):
            return list_of_object_names is None or object_name in list_of_object_names

        def load_if_exists(object_name, force=False):
            file_path = os.path.join(input_model_dir, f'{object_name}.pth')
            if os.path.exists(file_path):
                object_to_set = torch.load(file_path, map_location='cpu')
                loaded_objects.update({object_name: object_to_set})
                return object_to_set
            if force:
                self.logger.error(f"Cannot load this file: {file_path}")
                raise Exception(f"Cannot load this file: {file_path}")

        loaded_objects = {}

        if isinstance(input_model_dir, (tuple, list)):  # if multiple model dirs, load ensemble modules
            if mode != 'infer':
                msg = "Error in load_objects: using multiple model dirs is only supported in 'infer' mode"
                self.logger.error(msg)
                raise Exception(msg)

            # log an ensemble of model dirs
            self.logger.info('Loading ensemble of %d models.' % len(input_model_dir))
            for model_idx, model_dir in enumerate(input_model_dir):
                self.logger.info("Loading ensemble model %d from: %s" % (model_idx, model_dir))

            self.state.net = FuseModelEnsemble(input_model_dir)
            input_model_dir = input_model_dir[0]

        else:  # load single module
            self.logger.info("Loading model from: %s" % input_model_dir)
            self.state.net = load_if_exists('net', force=True) if should_load('net') else self.state.net

        # if mode is infer, then we need only load the net
        if mode == 'infer':
            load_if_exists('inference_dataset') if should_load('inference_dataset') else self.state.metrics
            return loaded_objects

        self.state.metrics = load_if_exists('metrics') if should_load('metrics') else self.state.metrics
        self.state.losses = load_if_exists('losses') if should_load('losses') else self.state.losses
        self.callbacks = load_if_exists('callbacks') if should_load('callbacks') else self.callbacks
        self.state.optimizer = load_if_exists('optimizer') if should_load('optimizer') else self.state.optimizer
        self.state.lr_scheduler = load_if_exists('lr_scheduler') if should_load('lr_scheduler') else self.state.lr_scheduler
        self.state.best_epoch_source = load_if_exists('best_epoch_source') if should_load('best_epoch_source') else self.state.best_epoch_source
        self.state.train_params = load_if_exists('train_params') if should_load('train_params') else self.state.train_params

        # train
        return loaded_objects

    def load_checkpoint(self,
                        checkpoint: Union[str, int, Sequence[Union[str, int]]],
                        model_dir: Optional[Union[str, Sequence[str]]] = None,
                        values_to_resume: List[str] = None,
                        mode: str = 'infer', strict: bool = True,
                        index: int = 0) -> None:
        """
        Loads values saved on checkpoint file.
        When mode = 'infer' loads only the model.

        :param checkpoint: checkpoint definition(s) to load weights from.
                           Possible values are 'best', 'last', epoch number (int) or full path to checkpoint file.
                           Can be a list, in case of an ensemble
        :param model_dir: source folder(s). might be ignored if checkpoint is a full path. Can be a list.
        :param values_to_resume:
            can contain any of: ['net', 'start_epoch', 'learning_rate'].
            when None (default) all values saved to checkpoint are loaded.
        :param mode: either 'infer' (default) or 'train'
        :param strict: input param to net.load_state_dict
        :param index: multiple checkpoint may be saved as best epoch, index indicates which one to take
        """

        def should_load(object_name):
            return values_to_resume is None or object_name in values_to_resume

        # possible values: epoch number, 'last', 'best', or full path to checkpoint file
        checkpoints = [checkpoint] if not isinstance(checkpoint, (list, tuple)) else checkpoint

        if model_dir is not None:
            model_dirs = [model_dir] if not isinstance(model_dir, (list, tuple)) else model_dir

            if len(model_dirs) > 1 and len(checkpoints) == 1:
                checkpoints = checkpoints * len(model_dirs)

        checkpoint_objs = []

        for idx in range(len(checkpoints)):
            checkpoint_desc = checkpoints[idx]
            # if checkpoint_desc is [int, 'best', 'last'], convert it to full path
            if isinstance(checkpoint_desc, int) or (isinstance(checkpoint_desc, str) and (checkpoint_desc in ['best', 'last'])):
                if isinstance(checkpoint_desc, str) and (checkpoint_desc == 'best'):
                    checkpoint_desc = checkpoint_desc + '_' + str(index)
                checkpoint_file = os.path.join(model_dirs[idx], f"checkpoint_{checkpoint_desc}_epoch.pth")
            # otherwise, checkpoint_desc is already a full path to checkpoint
            elif isinstance(checkpoint, str):
                checkpoint_file = checkpoint_desc
            else:
                msg = "Wrong checkpoint definition in 'load_checkpoint'. Possible values: integer/'best'/'last'/full path, but got: %s" % str(
                    checkpoint_desc)
                self.logger.error(msg)
                raise Exception(msg)
            str_vals = 'all' if values_to_resume is None else str(values_to_resume)
            self.logger.info(f'Loading checkpoint file: {checkpoint_file}. values_to_resume {str_vals}', {'color': 'yellow'})
            checkpoint_objs.append(FuseCheckpoint.load_from_file(checkpoint_file))

        if should_load('net'):
            net_state_dict_list = [checkpoint.net_state_dict for checkpoint in checkpoint_objs]
            self.state.net.load_state_dict(*net_state_dict_list, strict=strict)

        if mode == 'train':
            checkpoint_obj = checkpoint_objs[0]
            # on infer mode we don't need to load starting epoch or learning rate
            if should_load('start_epoch'):
                self.state.current_epoch = checkpoint_obj.epoch_idx
                self.logger.info(f'Loaded start epoch: {self.state.current_epoch}', {'color': 'yellow'})

            if should_load('learning_rate'):
                self.state.learning_rate = checkpoint_obj.learning_rate
                for param_group in self.state.optimizer.param_groups:
                    param_group['lr'] = self.state.learning_rate
                self.logger.info(f'Loaded learning rate: {self.state.learning_rate}', {'color': 'yellow'})

        pass

    def train(self, train_dataloader: DataLoader, validation_dataloader: DataLoader = None) -> None:
        """
        Train the net using train dataloader and validation on validation dataloader.

        :param train_dataloader: training data to use
        :param validation_dataloader: validation data to use.
            When None (default), validation is skipped and the best epoch values are decided by the train metrics.
        """
        # check that all objects needed are there

        self._verify_all_objects_initialized(mode='train')

        # debug - num workers
        override_num_workers = FuseUtilsDebug().get_setting('manager_override_num_dataloader_workers')
        if override_num_workers != 'default':
            train_dataloader.num_workers = override_num_workers
            validation_dataloader.num_workers = override_num_workers
            self.logger.info(f'Manager - debug mode - override dataloader num_workers to {override_num_workers}', {'color': 'red'})

        # prepare to use on GPUs
        self.state.net.to(self.state.device)
        self.state.net = nn.DataParallel(self.state.net)

        # TODO move losses to device as well
        total_param = sum(p.numel() for p in self.state.net.parameters())
        trainable_param = sum(p.numel() for p in self.state.net.parameters() if p.requires_grad)
        self.logger.info(f"Total number of parameters in model:{total_param:,}, trainable parameters:{trainable_param:,}",
                         {'color': 'red', 'attrs': 'bold'})

        # save model and parameters for future use (e.g., infer or resume_from_weights)
        self._save_objects(validation_dataloader)

        # save datasets summary into file and logger
        self._handle_dataset_summaries(train_dataloader, validation_dataloader)

        # handle callbacks
        for callback in self.callbacks: callback.on_train_begin(self.state)

        # validation handle_epoch, to see initial state of net
        if validation_dataloader is not None:
            initial_results = self.handle_epoch('validation', 0, validation_dataloader)
        else:
            initial_results = self.handle_epoch('validation', 0, train_dataloader)
        self.state.best_epoch_values = [initial_results for i in range(self.state.num_models_to_save)]
        self.state.current_epoch += 1

        # loop over num of epochs
        while self.state.current_epoch < self.state.end_epoch:
            for callback in self.callbacks: callback.on_step_begin(self.state.current_epoch)

            # train epoch
            self.logger.info(f"Start training on epoch {self.state.current_epoch}")
            train_results = self.handle_epoch('train', self.state.current_epoch, train_dataloader)

            # validation epoch - only if validation is needed
            if validation_dataloader is not None:
                self.logger.info(f"Start validation on epoch {self.state.current_epoch}")
                validation_results = self.handle_epoch('validation', self.state.current_epoch, validation_dataloader)
            else:
                validation_results = None

            epoch_checkpoint = FuseCheckpoint(self.state.net.module.state_dict(), self.state.current_epoch, self.get_current_learning_rate())

            # if this is the best epoch yet
            for i in range(self.state.num_models_to_save):
                if self._is_best_epoch_so_far(train_results, validation_results, i):
                    best_val = FuseUtilsHierarchicalDict.get(self.state.best_epoch_values[i], self.state.best_epoch_function[i])
                    self.logger.info(f"This is the best epoch ever ({self.state.best_epoch_function[i]} = {best_val})",
                                     {'color': 'green', 'attrs': 'bold'})
                    self.state.best_epoch[i] = self.state.current_epoch
                    best_epoch_checkpoint_filename = os.path.join(self.state.output_model_dir, 'checkpoint_best_' + str(i) + '_epoch.pth')
                    epoch_checkpoint.save_to_file(best_epoch_checkpoint_filename)
                # output to screen
                self._write_epoch_summary_table(train_results, validation_results, i)
            # save checkpoint to last epoch file
            last_epoch_checkpoint_filename = os.path.join(self.state.output_model_dir, 'checkpoint_last_epoch.pth')
            epoch_checkpoint.save_to_file(last_epoch_checkpoint_filename)

            if self.is_epoch_for_save(self.state.current_epoch):
                this_epoch_checkpoint_filename = os.path.join(self.state.output_model_dir, f'checkpoint_{self.state.current_epoch}_epoch.pth')
                epoch_checkpoint.save_to_file(this_epoch_checkpoint_filename)

            # LR scheduler update and log
            self.update_scheduler(train_results, validation_results)

            for callback in self.callbacks:
                callback.on_step_end(self.state.current_epoch, train_results, validation_results, self.get_current_learning_rate())

            self.state.current_epoch += 1

        for callback in self.callbacks: callback.on_train_end()

        pass

    def visualize(self, visualizer: FuseVisualizerBase, data_loader: Optional[DataLoader] = None, infer_processor: Optional[FuseProcessorBase] = None,
                  descriptors: Optional[List[Hashable]] = None, device: str = 'cuda', display_func: Optional[Callable] = None):

        """
        Visualize data including the input and the output.
        Expected Sequence:
        1. Using a loaded model to extract the output:
         manager = FuseManagerDefault()

         manager.load_objects(<model dir>, mode='infer')  # this method can load either a single model or an ensemble
         manager.load_checkpoint(checkpoint=<path to checkpoint file>, mode='infer')
         manager.visualize(visualizer=visualizer,
                  data_loader=dataloader,
                  descriptors=<optional - list of descriptors>,
                  display_func=<optional - display_func>,
                  infer_processor=None)

        2. using inference processor
         manager = FuseManagerDefault()
         manager.visualize(visualizer=visualizer,
                  data_loader=dataloader,
                  descriptors=<optional - list of descriptors>,
                  display_func=<optional - display_func>,
                  infer_processor=infer_processor)

        :param visualizer: The visualizer, getting a batch_dict as an input and doing it's magic
        :param data_loader: data loader as used for validation / training / inference
        :param infer_processor: Optional, if specified this function will not run the model and instead extract the output from infer processor
        :param descriptors: Optional. List of sample descriptors, if None will go over the entire dataset. Might be also list of dataset indices.
        :param device: options: 'cuda', 'cpu', 'cuda:0', ... (default 'cuda')
        :param display_func: Function getting the batch dict as an input and returns boolean specifying if to visualize this sample or not.
        :return: None
        """
        dataset: FuseDatasetBase = data_loader.dataset
        if infer_processor is None:
            if not hasattr(self, 'net') or self.state.net is None:
                self.logger.error(f"Cannot visualize without either net or infer_processor")
                raise Exception(f"Cannot visualize without either net or infer_processor")

            # prepare net
            self.state.net.to(device)
            self.state.net = nn.DataParallel(self.state.net)

        if descriptors is None:
            descriptors = range(len(dataset))
        for desc in tqdm(descriptors):
            # extract sample
            batch_dict = dataset.get(desc)
            if infer_processor is None:
                # apply model in case infer processor is not specified
                # convert dimensions to batch
                batch_dict = dataset.collate_fn([batch_dict])
                # run model
                batch_dict['model'] = self.state.net(batch_dict)
                # convert dimensions back to single sample
                FuseUtilsHierarchicalDict.apply_on_all(batch_dict, FuseUtilsMisc.squeeze_obj)
            else:
                # get the sample descriptor of the sample
                sample_descriptor = FuseUtilsHierarchicalDict.get(batch_dict, 'data.descriptor')
                # get the infer data
                infer_data = infer_processor(sample_descriptor)
                # add infer data to batch_dict
                for key in infer_data:
                    FuseUtilsHierarchicalDict.set(batch_dict, key, infer_data[key])

            if display_func is None or display_func(batch_dict):
                visualizer.visualize(batch_dict)

    def infer(self, input_model_dir: Optional[Union[str, Sequence]] = None,
              checkpoint: Optional[Union[str, int, Sequence[Union[str, int]]]] = None,
              data_source: Optional[FuseDataSourceBase] = None, data_loader: Optional[DataLoader] = None,
              num_workers: Optional[int] = 4, batch_size: Optional[int] = 2,
              output_columns: List[str] = None, output_file_name: str = None, strict: bool = True,
              append_default_inference_callback: bool = True,
              checkpoint_index: int = 0) -> pd.DataFrame:
        """
        Inference of net on data. Either the data_source or data_loader should be defined.
        When data_source is defined, validation_dataset is loaded from the original model_dir and is used to create a dataloader.
        Returns the inference Results as dict:
                {
                'descriptor': [id_1, id_2, ...],
                'output': {'output_head1': [res1_1, res1_2, ...],
                           'output_head2': [res2_1, res2_2, ..]}
                }

        ================================================================================================================
         Ensemble Mode
        ===============
        Optionally, run inference using an ensemble of several models. This is done by providing a list of
        model dirs in the 'input_model_dir' param and optionally a list of checkpoints in the 'checkpoint' param.
        When using an ensemble, the first model dir is used for loading Fuse data pipeline: Dataset and Input Processors.
        All model dirs, including the first, are used for loading modules and their checkpoints for the ensemble.
        In this mode, inference Results dict will look like:
                {
                'descriptor': [id_1, id_2, ...],
                'ensemble_output_0': {'output_head1': [res1_1, res1_2, ...],     # first ensemble model
                                      'output_head2': [res2_1, res2_2, ..]},
                'ensemble_output_1': {'output_head1': [res1_1, res1_2, ...],     # second ensemble model
                                      'output_head2': [res2_1, res2_2, ..]},
                ...

                }
        ================================================================================================================

        :param input_model_dir: model directory(s) to load data from
        :param checkpoint: checkpoint definition(s) to load weights from. Possible values are 'best', 'last' or int.
            When not None, model weights are loaded from respected checkpoint file under input_model_dir
            (either checkpoint_best_epoch.pth, checkpoint_last_epoch.pth or checkpoint_{checkpoint}_epoch.pth)
            when None, no checkpoint is loaded (assumes that the weights were already loaded.
            in ensemble mode, can provide either one checkpoint for all models or a sequence of separate checkpoints for each.
        :param data_source: data source to use
        :param data_loader: data loader to use
        :param num_workers: number of processes for Dataloader, effective only if 'data_loader' param is None
        :param batch_size: batch size for Dataloader, effective only if 'data_loader' param is None
        :param output_columns: output columns to return.
            When None (default) all columns are returned.
            When not None, FuseInferResultsCallback callback is created.
        :param output_file_name: output file path. when None (default) results are not saved to file.
        :param strict: strict state dict loading when loading checkpoint weights. default is True.
        :param append_default_inference_callback: if True, appends Fuse's default results collector callback
        :param checkpoint_index: few best checkpoints can be saved, each with its own index
        :return: infer results in a DataFrame
        """

        # debug - num workers
        override_num_workers = FuseUtilsDebug().get_setting('manager_override_num_dataloader_workers')
        if override_num_workers != 'default':
            num_workers = override_num_workers
            if data_loader is not None:
                data_loader.num_workers = override_num_workers
            self.logger.info(f'Manager - debug mode - override dataloader num_workers to {override_num_workers}', {'color': 'red'})

        if input_model_dir is not None:
            # user provided model dir(s), and Manager has no 'net' attribute - need to load modules
            if not hasattr(self.state, 'net'):
                self.load_objects(input_model_dir, mode='infer', list_of_object_names=['net'])  # this method can load either a single model or an ensemble

        if checkpoint is not None:
            if hasattr(self.state, 'net'):
                if isinstance(self.state.net, torch.nn.DataParallel):
                    raise Exception("Error in infer - Manager has a DataParallel net. Cannot load checkpoint into DataParallel module!")
                if input_model_dir is None:
                    msg = "Cannot load checkpoint file without a definition of input_model_dir"
                    self.logger.error(msg)
                    raise Exception(msg)

                self.load_checkpoint(checkpoint=checkpoint, model_dir=input_model_dir, mode='infer', strict=strict, index=checkpoint_index)

        # checklist check that all objects needed are there
        self._verify_all_objects_initialized(mode='infer')

        #TODO I don't like this flag - maybe think about a way to get rid of it?
        # append inference callback
        if append_default_inference_callback:
            self.callbacks.append(FuseInferResultsCallback(output_file=output_file_name, output_columns=output_columns))

        # either optional_datasource or optional_dataloader
        if data_loader is not None and data_source is not None:
            self.logger.error('Cannot have both data_loader and data_source defined')
            raise Exception('Cannot have both data_loader and data_source defined')
        if data_loader is None and data_source is None:
            self.logger.error('Either data_loader or data_source should be defined')
            raise Exception('Either data_loader or data_source should be defined')

        if data_loader is None:
            # need to create a data loader
            # first check that we have the model dir to get these data from
            if input_model_dir is None:
                self.logger.error('Missing parameter input_model_dir! Cannot load data_set from previous model.')
                raise Exception('Missing parameter input_model_dir! Cannot load data_set from previous model.')

            if isinstance(input_model_dir, (tuple, list)):
                data_set_filename = os.path.join(input_model_dir[0], "inference_dataset.pth")
            else:
                data_set_filename = os.path.join(input_model_dir, "inference_dataset.pth")
            self.logger.info(f"Loading data source definitions from {data_set_filename}", {'color': 'yellow'})
            infer_dataset = FuseDatasetBase.load(filename=data_set_filename, override_datasource=data_source)
            data_loader = DataLoader(dataset=infer_dataset, shuffle=False, drop_last=False, batch_sampler=None,
                                     batch_size=batch_size, num_workers=num_workers, collate_fn=infer_dataset.collate_fn)

        # prepare net
        self.state.net.to(self.state.device)
        self.state.net = nn.DataParallel(self.state.net)

        # we are ready to run inference
        self.handle_epoch('infer', 0, data_loader)

        # if infer CB is in the callback list, then return its result
        for callback in self.callbacks:
            if isinstance(callback, FuseInferResultsCallback):
                return callback.get_infer_results()

    def handle_epoch(self, mode: str, epoch: int, data_loader: DataLoader) -> Dict:
        """
        For each virtual batch calls handle_virtual_batch() and aggregates the results into a dict.
        In case mode is 'infer', the prediction results are returned.
        in case mode is 'train' or 'validation', the losses are returned.
        :param mode: mode to run the epoch. Can be either 'train', 'validation', 'infer'
        :param epoch: epoch number
        :param data_loader: data to load
        :return: a hierarchical dictionary depending on the epoch mode.
            For infer:
                epoch_results = {
                            'descriptor': [id_1, id_2, ...],
                            'output': {'output_head1': [res1_1, res1_2, ...],
                                        'output_head2': [res2_1, res2_2, ..]}
                            }
            For Train or Validation:
               epoch_results = {
                            'losses': {'loss1': mean_loss1,
                            '          'loss2': mean_loss2},
                            'metrics': {'metric1': epoch_metric1,
                                        'metric2': epoch_metric2}
               }
        """

        for callback in self.callbacks: callback.on_epoch_begin(mode=mode, epoch=epoch)
        assert mode in ['train', 'validation', 'infer']

        if mode == 'train':
            with torch.enable_grad():
                self.state.net.train()
                epoch_results = self.do_handle_epoch(mode, epoch, data_loader)
                for callback in self.callbacks:
                    callback.on_epoch_end(mode, epoch, epoch_results)
                return epoch_results
        elif mode in ['validation', 'infer']:
            with torch.no_grad():
                self.state.net.eval()
                epoch_results = self.do_handle_epoch(mode, epoch, data_loader)
                for callback in self.callbacks:
                    callback.on_epoch_end(mode, epoch, epoch_results)
                return epoch_results

    def do_handle_epoch(self, mode: str, epoch: int, data_loader: DataLoader) -> Dict:
        """
        Utility function that is called by handle_epoch
        :param mode: mode to run the epoch. Can be either 'train', 'validation', 'infer'
        :param epoch: epoch number
        :param data_loader: data to load
        :return: a dictionary depending on the epoch mode.
            For Train or Validation:
               epoch_results = {

               epoch_results = {
                            'losses': {'loss1': mean_loss1,
                                      'loss2': mean_loss2,
                                      'total_loss': mean_loss1 + mean_loss2},
                            'metrics': {'metric1': epoch_metric1,
                                        'metric2': epoch_metric2}
                                                       }
            For infer returns an empty dict {}.
        """
        # handle each virtual batch separately
        num_batches = int(np.ceil(len(data_loader) / self.state.virtual_batch_size))
        data_iter = iter(data_loader)

        # loop over batches (can be virtual batch)
        epoch_results = {}
        for virtual_batch in trange(num_batches):
            # handle_virtual batch
            virtual_batch_dict = self.handle_virtual_batch(mode, virtual_batch, data_iter)
            epoch_results = _extend_results_dict(mode, virtual_batch_dict, epoch_results)

        # compute metrics and keep the results
        for metric_name, metric in self.state.metrics.items():
            try:
                metric_result = metric.process()
            except:
                lgr = logging.getLogger('Fuse')
                track = traceback.format_exc()
                lgr.error(f'Metric {metric_name} process() func failed. Setting results to None')
                lgr.error(track)
                metric_result = None

            FuseUtilsHierarchicalDict.set(epoch_results, "metrics." + metric_name, metric_result)
            metric.reset()

        # average losses into mean_loss
        if 'losses' in epoch_results:
            for loss in FuseUtilsHierarchicalDict.get_all_keys(epoch_results['losses']):
                batch_losses = FuseUtilsHierarchicalDict.get(epoch_results, 'losses.' + loss)
                loss_mean = np.nansum(batch_losses) / len(batch_losses)
                FuseUtilsHierarchicalDict.set(epoch_results, "losses." + loss, loss_mean)

        return epoch_results

    def handle_virtual_batch(self, mode: str, virtual_batch: int, data_iter: Iterator) -> Dict:
        """
        Responsible for splitting into mini batches, running each mini batch, aggregating its results, and running optimizer at the end

        :param mode: mode to run the epoch. Can be either 'train', 'validation', 'infer'
        :param virtual_batch: virtual batch number
        :param data_iter: iterator of data to use
        :return: a hierarchical dictionary with data of virtual batch, depending on the mode.
            For infer:
                epoch_results = {
                            'descriptor': [id_1, id_2, ...],
                            'output': {'output_head1': [res1_1, res1_2, ...],
                                        'output_head2': [res2_1, res2_2, ..]}
                            }
            For Train or Validation:
               epoch_results = {
                            'losses' : {'loss1': [0.21, 0.45362],
                                        'loss2': [0.01, 0.443],
                                        'total_loss': [0.22, 0.89662]
                                        }
               }
        """
        for callback in self.callbacks: callback.on_virtual_batch_begin(mode=mode, virtual_batch=virtual_batch)

        # mode is train/validation/infer
        if mode == 'train':
            self.state.optimizer.zero_grad()

        virtual_batch_results = {}

        for mini_batch in range(self.state.virtual_batch_size):
            mini_batch_result_dict = self.handle_batch(mode, mini_batch, data_iter)
            virtual_batch_results = _extend_results_dict(mode, mini_batch_result_dict, virtual_batch_results)

        if mode == 'train':
            # after all virtual mini batches all processed, we can run the optimizer
            self.state.optimizer.step(closure=self.state.opt_closure)

        for callback in self.callbacks: callback.on_virtual_batch_end(mode, virtual_batch, virtual_batch_results)

        return virtual_batch_results

    def handle_batch(self, mode: str, batch: int, data_iter: Iterator) -> Dict:
        """Handles the batch load, net forward and metrics computations
        :param mode: mode to run the epoch. Can be either 'train', 'validation', 'infer'
        :param batch: batch number
        :param data_iter: data to load
        :return: hierarchical batch_dict containing:
            descriptor: unique identifier for each sample processed,
            model.output: a dict with keys for each possible output,
            losses: a dict with a key for each defined loss + a total_loss key
        """
        # callbacks handling
        for callback in self.callbacks: callback.on_batch_begin(mode, batch)

        # get the input
        try:
            batch_dict = next(data_iter)
        # in case this was called from the last virtual batch, and we don't have any more inputs
        except StopIteration:
            # callbacks handling
            for callback in self.callbacks: callback.on_batch_end(mode, batch, {})
            return {}

        for callback in self.callbacks: callback.on_data_fetch_end(mode, batch, batch_dict)

        # move every tensor in input to device
        FuseUtilsHierarchicalDict.apply_on_all(batch_dict, gpu.move_tensor_to_device, self.state.device)

        # forward net
        batch_dict['model'] = self.state.net(batch_dict)

        # compute total loss and keep loss results
        total_loss: torch.Tensor = 0
        for loss_name, loss_function in self.state.losses.items():
            current_loss_result = loss_function(batch_dict)
            FuseUtilsHierarchicalDict.set(batch_dict, 'losses.' + loss_name, current_loss_result.data.item())
            # sum all losses for backward
            total_loss += current_loss_result
        # no need to add total_loss if there are no losses computed
        if isinstance(total_loss, torch.Tensor):
            FuseUtilsHierarchicalDict.set(batch_dict, 'losses.total_loss', total_loss.data.item())

        if mode == 'train':
            # backward
            total_loss.backward()

        for callback in self.callbacks: callback.on_batch_end(mode, batch, batch_dict=batch_dict)

        # compute metrics
        for metric_name, metric in self.state.metrics.items():
            # handle batch doesn't return a value, the actual value of the metric is per epoch
            metric.collect(batch_dict)


        return batch_dict

    def update_scheduler(self, train_results: Dict, validation_results: Dict) -> None:
        """
        Update the scheduler using the input parameters. this function is called at the end of each epoch run.
        If the learning rate has changed during the step function, the log is updated accordingly.

        :param train_results: hierarchical dict train epoch results.
            contains the keys: losses, metrics.
                losses is a dict where values are the commputed mean loss for each loss.
                    and an additional key 'total_loss' which is the mean total loss of the epoch.
                metrics is a dict where values are the computed metrics.
        :param validation_results: hierarchical validation epoch results dict.
             contains the keys: losses, metrics.
                losses is a dict where values are the commputed mean loss for each loss.
                    and an additional key 'total_loss' which is the mean total loss of the epoch.
                metrics is a dict where values are the computed metrics.
                Note, if validation was not done on the epoch, this parameter can be None
        """
        # implementation for the train ReduceLROnPlateau scheduler:
        prev_lr = self.get_current_learning_rate()

        # take total_loss from train_results
        self.state.lr_scheduler.step(np.mean(FuseUtilsHierarchicalDict.get(train_results, 'losses.total_loss')))

        curr_lr = self.get_current_learning_rate()

        # log scheduler changes - if lr has changed during step
        if prev_lr != curr_lr:
            self.logger.info(f"Learning rate has changed from {prev_lr} to {curr_lr}")

    def _verify_all_objects_initialized(self, mode: str):
        def verify_value(self_object, parameter):
            if self_object is None:
                self.logger.error(f"Cannot {mode} without {parameter} definition")
                raise Exception(f"Cannot {mode} without {parameter} definition")

        verify_value(self.state.net, 'net')
        verify_value(self.callbacks, 'callbacks')

        if mode == 'train':
            verify_value(self.state.metrics, 'metrics')
            verify_value(self.state.losses, 'losses')
            verify_value(self.state.optimizer, 'optimizer')
            verify_value(self.state.lr_scheduler, 'lr_scheduler')
            verify_value(self.state.train_params, 'train_params')
            verify_value(self.state.best_epoch_source, 'best_epoch_source')
            verify_value(self.state.output_model_dir, 'output_model_dir')

        # make sure we have all the parameters we need
        full_config = self.set_config_defaults(self.state.train_params, mode)

        self.state.virtual_batch_size: int = full_config['virtual_batch_size']

        if mode == 'train':
            self.state.num_epochs: int = full_config['num_epochs']
            # debug - num epochs
            override_num_epochs = FuseUtilsDebug().get_setting('manager_override_num_epochs')
            if override_num_epochs != 'default':
                self.state.num_epochs = override_num_epochs
                self.logger.info(f'Manager - debug mode - override num_epochs to {self.state.num_epochs}', {'color': 'red'})

            self.state.start_saving_epochs: int = full_config['start_saving_epochs']
            self.state.gap_between_saving_epochs: int = full_config['gap_between_saving_epochs']
            self.state.end_epoch: int = self.state.num_epochs

            self.state.num_models_to_save = 1 if isinstance(self.state.best_epoch_source, dict) else len(self.state.best_epoch_source)
            self.state.best_epoch = [0 for _ in range(self.state.num_models_to_save)]
            for i in range(self.state.num_models_to_save):
                # update best epoch source members
                if self.state.metrics:
                    self.state.best_epoch_function.append(
                        self.state.best_epoch_source[i].get('source', f'metrics.{list(self.state.metrics.keys())[0]}'))
                else:
                    self.state.best_epoch_function.append(
                        self.state.best_epoch_source[i].get('source', f'losses.{list(self.state.losses.keys())[0]}'))

                self.state.optimization_function.append(self.state.best_epoch_source[i].get('optimization', 'max'))
                assert self.state.optimization_function[i] in ['max', 'min']
                initial_value = -float('inf') if self.state.optimization_function[i] == 'max' else float('inf')
                self.state.best_epoch_values.append({self.state.best_epoch_function[i]: initial_value})

                on_equal_values = self.state.best_epoch_source[i].get('on_equal_values', 'better')
                assert on_equal_values in ['better', 'worse']
                self.state.on_equal_values.append(on_equal_values)

        self.state.device: str = full_config.get('device')
        pass

    def is_epoch_for_save(self, epoch: int) -> bool:
        """
        Checks whether this epoch should be saved according to the save epoch parameters
        :param epoch: epoch num
        :return: True if the epoch checkpoint should be saved, False Otherwise.
        """

        # save every gap_between_saving_epochs epoch starting from the start_saving_epochs epoch
        return epoch >= self.state.start_saving_epochs and (epoch - self.state.start_saving_epochs) % self.state.gap_between_saving_epochs == 0

    def _is_best_epoch_so_far(self, train_results: Dict, validation_results: Dict, epoch_source_index: int) -> bool:
        """
        Returns true whether the current results are the best results by now.
        if validation results are None, use train_results to decide.
        Uses the definitions of self.state.best_epoch_source in order to decide.
        :param train_results: train results
        :param validation_results:  validation results
        :return: True if results are best so far.
        """

        def is_better_epoch_value(current_value: float) -> bool:
            try:
                current_best = FuseUtilsHierarchicalDict.get(self.state.best_epoch_values[epoch_source_index],
                                                             self.state.best_epoch_function[epoch_source_index])
                # check if we can compare values
                if current_value is None:
                    lgr = logging.getLogger('Fuse')
                    lgr.error(f'Comparing epochs failed since value is None, assuming it is not the best epoch')
                    lgr.error(traceback.format_exc())
                    return False

                # first handle equal values
                if current_value == current_best:
                    return self.state.on_equal_values[epoch_source_index] == 'better'

                # for different values: if max is better
                if self.state.optimization_function[epoch_source_index] == 'max':
                    return current_value > current_best
                # if min is better
                return current_value < current_best
            except:
                lgr = logging.getLogger('Fuse')
                track = traceback.format_exc()
                lgr.error(f'Comparing epochs failed, assuming it is not the best epoch')
                lgr.error(track)
                return False

        if validation_results is not None:
            values_to_check = validation_results
        else:
            values_to_check = train_results
        function_key = self.state.best_epoch_function[epoch_source_index]
        if not FuseUtilsHierarchicalDict.is_in(values_to_check, function_key):
            lgr = logging.getLogger('Fuse')
            lgr.error(f"source function {function_key} does not exist in results_dict. " + \
                      f"Possible values are {FuseUtilsHierarchicalDict.get_all_keys(values_to_check)}")
            lgr.error(traceback.format_exc())
            raise KeyError(f"source function {function_key} does not exist in results_dict." + \
                           f"Possible values are {FuseUtilsHierarchicalDict.get_all_keys(values_to_check)}")

        value_to_compare = FuseUtilsHierarchicalDict.get(values_to_check, function_key)
        if is_better_epoch_value(value_to_compare):
            self.state.best_epoch_values[epoch_source_index] = values_to_check
            return True
        return False

    def get_current_learning_rate(self) -> float:
        """
        Returns the current learning rate used by the optimizer
        :return: learning rate
        """
        return self.state.optimizer.param_groups[0]['lr']

    def _write_epoch_summary_table(self, train_dict: dict, validation_dict: dict, epoch_source_index: int) -> None:
        def get_value_as_float_str(dict, key):
            val_as_str = 'N/A'
            try:
                value = FuseUtilsHierarchicalDict.get(dict, key)
                val_as_str = '%.4f' % float(value)
            except:
                pass
            return val_as_str

        stats_table = pd.DataFrame(columns=['', 'Best Epoch Value', 'Current Epoch Validation', 'Current Epoch Train'])
        idx = 0

        eval_keys = sorted(FuseUtilsHierarchicalDict.get_all_keys(train_dict))
        for evaluator_name in eval_keys:
            train_value_str = get_value_as_float_str(train_dict, evaluator_name)
            validation_val_str = get_value_as_float_str(validation_dict, evaluator_name)
            best_so_far_str = get_value_as_float_str(self.state.best_epoch_values[epoch_source_index], evaluator_name)

            stats_table.loc[idx] = [f'{evaluator_name}', best_so_far_str, validation_val_str, train_value_str]
            idx += 1

        best_source = self.state.best_epoch_function[epoch_source_index]
        if self.state.current_epoch == self.state.best_epoch[epoch_source_index]:
            epoch_title = f'Stats for epoch: {self.state.current_epoch} (Currently the best epoch for source {best_source}!)'
            self.logger.info(epoch_title, {'attrs': ['underline', 'bold']})
        else:
            best_so_far = self.state.best_epoch[epoch_source_index]
            epoch_title = f'Stats for epoch: {self.state.current_epoch} (Best epoch is {best_so_far} for source {best_source})'
            self.logger.info(epoch_title)
        stats_as_str = misc.get_pretty_dataframe(stats_table)
        self.logger.info(stats_as_str)
        self.logger.info(f"Model: {self.state.output_model_dir}")

        try:
            op = 'w' if epoch_source_index == 0 else 'a'  # we want to have all stats in one file per epoch
            with open(os.path.join(self.state.output_model_dir, "last_epoch_summary.txt"), op) as sfile:
                sfile.write(epoch_title)
                sfile.write(stats_as_str)
        except Exception as error:
            self.logger.error(f"Cannot write epoch summary to file {self.state.output_model_dir}/last_epoch_summary.txt")
            self.logger.error(error)
        pass

    def set_config_defaults(self, config: dict, mode: str) -> Dict:
        """
        Sets the train_param default values (if not defined). log in case of updating a key.
        :param config: train_config
        :param mode: either 'infer' or 'train' (in case of infer, only device and virtual_batch_size are set)
        :return: dict containing all expected values
        """

        def set_default(key, value):
            if key not in full_config:
                full_config[key] = value
                self.logger.info(f"Key {key} not found in config parameter, setting value to default ({value})", {'color': 'yellow'})

        full_config = {} if config is None else config.copy()
        set_default('device', 'cuda')
        set_default('virtual_batch_size', 1)

        if mode == 'train':
            set_default('num_epochs', 100)
            set_default('gap_between_saving_epochs', 5)
            set_default('start_saving_epochs', 80)

        return full_config

    def _handle_dataset_summaries(self, train_dataloader: DataLoader, validation_dataloader: DataLoader) -> None:
        """
        Write summary of datasets into files and logger
        :param train_dataloader: train data
        :param validation_dataloader: validation data (can be None)
        """
        # train dataset summary
        dataset_summary = train_dataloader.dataset.summary()

        train_dataset_summary_file = os.path.join(self.state.output_model_dir, 'train_dataset_summary.txt')
        with open(train_dataset_summary_file, 'w') as sum_file:
            sum_file.write(dataset_summary)
        self.logger.info("Train Dataset Summary:")
        self.logger.info(dataset_summary)

        # validation dataset summary, if exists
        if validation_dataloader is not None:
            dataset_summary = validation_dataloader.dataset.summary()
            validation_dataset_summary_file = os.path.join(self.state.output_model_dir, 'validation_dataset_summary.txt')
            with open(validation_dataset_summary_file, 'w') as sum_file:
                sum_file.write(dataset_summary)
            self.logger.info("Validation Dataset Summary:")
            self.logger.info(dataset_summary)
        pass


def _extend_results_dict(mode: str, current_dict: Dict, aggregated_dict: Dict) -> Dict:
    """
    Utility function to create aggregated loss results dict from the handle_batch()/handle_virtual_batch() output dict.
    :param mode: Can be either 'train', 'validation', 'infer' (on infer return an empty dict)
    :param current_dict: batch_dict or virtual_batch_result dict
    :param aggregated_dict: the aggregated dict to add current_dict
    :return: aggregated dict
    """
    if mode == 'infer':
        return {}
    else:
        # handle the case where batch dict is empty (the end of the last virtual mini batch)
        if current_dict == {}:
            return aggregated_dict
        # for train and validation we need the loss values
        cur_keys = FuseUtilsHierarchicalDict.get_all_keys(current_dict)
        # aggregate just keys that start with losses
        cur_keys = [key for key in cur_keys if key.startswith('losses.')]
        agg_keys = FuseUtilsHierarchicalDict.get_all_keys(aggregated_dict)
        for key in cur_keys:
            if key not in agg_keys:
                # init dict is needed
                FuseUtilsHierarchicalDict.set(aggregated_dict, key, [])
            val = FuseUtilsHierarchicalDict.get(current_dict, key)
            if isinstance(val, list):
                # in the epoch dict, the loss is a list of the virtual mini batches
                FuseUtilsHierarchicalDict.get(aggregated_dict, key).extend(val)
            else:
                # in the virtual batch dict, the losses are float per batch
                FuseUtilsHierarchicalDict.get(aggregated_dict, key).append(val)

    return aggregated_dict



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Sequence, Hashable

from fuse.data.data_source.data_source_base import FuseDataSourceBase


class FuseDataSourceFromList(FuseDataSourceBase):
    """
    Simple DataSource that can be initialized with a Python list (or other sequence).
    Does nothing but passing the list to Dataset.
    """

    def __init__(self, list_of_samples: Sequence[Hashable] = []) -> None:
        self.list_of_samples = list_of_samples

    def get_samples_description(self):
        return self.list_of_samples

    def summary(self) -> str:
        summary_str = ''
        summary_str += 'FuseDataSourceFromList - %d samples\n' % len(self.list_of_samples)
        return summary_str



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Dict, List, Tuple, Callable
import numpy as np
import pandas as pd

from Fuse.metrics.metric_base import FuseMetricBase
from Fuse.metrics.metrics_toolbox import FuseMetricsToolBox


class FuseMetricPredictionBreakdown(FuseMetricBase):
    """
    Multi class version for to breakdown prediction to subgroups/subtypes.

    """

    def __init__(self, pred_name: str, target_name: str, breakdown_class_name: str,
                 class_names: List = None, class_thresholds: List[Tuple] = None, **kwargs):
        """
        :param pred_name:       batch_dict key for predicted output (e.g., class probabilities after softmax)
        :param target_name:     batch_dict key for target (e.g., ground truth label)
        :param class_names:     Optional - name for each class otherwise the class index will be used.
        :param breakdown_class_name: sybtype of each sample,the breakdown will be according to this lable.
        :param class_thresholds: Optional - threshold for each class, the order of the list is the order in which class
                            is assigned, Each element in the list is a tuple of (class_idx, class_threshold)
                            if None: the class is set by argmax
        """
        super().__init__(pred_name, target_name, breakdown_class_name=breakdown_class_name, **kwargs)
        self._class_names = class_names
        self._class_thresholds = class_thresholds

    def process(self) -> Dict[str, float]:
        """
        Returns a string table with class prediction breakdown according to sample subtype
        """

        epoch_preds = np.array(self.epoch_preds)
        epoch_targets = np.array(self.epoch_targets)
        epoch_breakdown = np.array(self.collected_data['breakdown_class'])

        # first find operation points
        if isinstance(self._class_thresholds, Callable):
            class_thresholds = self._class_thresholds(epoch_preds, epoch_targets)
        else:
            class_thresholds = self._class_thresholds

        # get class predictions
        epoch_class_preds = FuseMetricsToolBox.convert_probabilities_to_class(epoch_preds, thresholds=class_thresholds)

        # get class names
        num_classes = epoch_preds[0].shape[0]
        class_names = FuseMetricsToolBox.get_class_names(num_classes, self._class_names)

        # get breakdown classes (sub types)
        breakdown_classes = sorted(list(set(epoch_breakdown)), key=lambda x: str(x))
        breakdown_classes_count = self.get_list_type_count(breakdown_classes, epoch_breakdown)

        # fill in a 2D table of breakdown_cls x predicted cls_
        breakdown_count_per_class = {str(cls): [] for cls in breakdown_classes}
        for breakdown_cls in breakdown_classes:
            breakdown_cls_name = str(breakdown_cls)
            for class_idx, class_name in enumerate(class_names):
                detect_count = np.logical_and(np.where(epoch_breakdown == breakdown_cls, 1, 0),
                                              np.where(epoch_class_preds == class_idx, 1, 0)).sum()
                all_count = breakdown_classes_count[breakdown_cls]
                count_percent = '(' + str(round(detect_count / int(all_count) * 100, 1)) + '%)'
                breakdown_count_per_class[breakdown_cls_name].append(str(detect_count) + '/' + all_count + count_percent)

        return pd.DataFrame(breakdown_count_per_class, index=class_names).transpose().to_string()

    @staticmethod
    def get_list_type_count(breakdown_classes, epoch_breakdown) -> Dict:
        return {breakdown_cls: str(list(epoch_breakdown).count(breakdown_cls)) for breakdown_cls in breakdown_classes}



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import logging
from typing import Optional, Iterable, Any, Tuple

import matplotlib.pyplot as plt
from skimage.color import gray2rgb
from skimage.segmentation import mark_boundaries

from fuse.data.visualizer.visualizer_base import FuseVisualizerBase
from fuse.utils.utils_hierarchical_dict import FuseUtilsHierarchicalDict
from fuse.utils.utils_logger import log_object_input_state


class Fuse3DVisualizerDefault(FuseVisualizerBase):
    """
    Visualiser for data including 3D volume with optional local annotations
    """

    def __init__(self, image_name: str, mask_name: Optional[str] = None,
                 label_name: Optional[str] = None, metadata_pointers: Iterable[str] = tuple(),
                 ):
        """
        :param image_name: pointer to an image in batch_dict, image will be in shape (B,C,VOL).
        :param mask_name:  optional, pointer mask (gt map) in batch_dict. If mask location is not part of the batch dict -
                            override the extract_data method
        :param label_name: pointer to a global label in batch_dict
        :param metadata_pointers: list of pointers to metadata - will be printed for every sample

        """
        # log object input state
        log_object_input_state(self, locals())

        # store input parameters
        self.image_name = image_name
        self.mask_name = mask_name
        self.label_name = label_name
        self.metadata_pointers = metadata_pointers

    def extract_data(self, sample: dict) -> Tuple[Any, Any, Any, Any]:
        """
        extract required data to visualize from sample
        :param sample: global dict of a sample
        :return: image, mask, label, metadata
        """

        # image
        image = FuseUtilsHierarchicalDict.get(sample, self.image_name)
        assert len(image.shape) == 4
        image = image.numpy()

        # mask
        if self.mask_name is not None:
            if not isinstance(self.mask_name, list):
                self.mask_name = [self.mask_name]
            masks = [FuseUtilsHierarchicalDict.get(sample, mask_name).numpy() for mask_name in self.mask_name]
        else:
            masks = None

        # label
        if self.label_name is not None:
            label = FuseUtilsHierarchicalDict.get(sample, self.label_name)
        else:
            label = ''

        # metadata
        metadata = {metadata_ptr: FuseUtilsHierarchicalDict.get(sample, metadata_ptr) for metadata_ptr in
                    self.metadata_pointers}

        return image, masks, label, metadata

    def visualize(self, sample: dict, block: bool = True) -> None:
        """
        visualize sample
        :param sample: batch_dict - to extract the sample from
        :param block: set to False if the process should not be blocked until the plot will be closed
        :return: None
        """
        # extract data
        image, masks, label, metadata = self.extract_data(sample)
        # visualize
        chan = 0
        chan_image = image[chan, ...]

        def key_event(e: Any, position_list: Any):  # using left/right key to move between slices
            def on_press(e: Any):  # use mouse click in order to toggle mask/no mask
                'toggle the visible state of the two images'
                if e.button:
                    vis_image = plt_img.get_visible()
                    vis_mask = plt_mask.get_visible()
                    plt_img.set_visible(not vis_image)
                    plt_mask.set_visible(not vis_mask)
                    plt.draw()

            if e.key == "right":
                position_list[0] += 1
            elif e.key == "left":
                position_list[0] -= 1
            elif e.key == "up":
                position_list[1] += 1
            elif e.key == "down":
                position_list[1] -= 1
            else:
                return
            position_list[0] = position_list[0] % image.shape[1]
            position_list[1] = position_list[1] % image.shape[0]
            chan_image = image[position_list[1]]

            ax.cla()
            slice_image = gray2rgb(chan_image[position_list[0]])
            if masks is not None:
                slice_image_with_mask = slice_image
                colors = [(1, 1, 0), (0, 1, 1), (1, 0, 1)]
                for index, mask in enumerate(masks):
                    slice_image_with_mask = mark_boundaries(slice_image_with_mask, mask[position_list[0]].astype(int),
                                                            color=colors[index % len(colors)])

            plt.title(f'Slice {position_list[0]} channel {position_list[1]}')
            plt_img = ax.imshow(slice_image)
            plt_img.set_visible(False)
            if (mask is not None) and (None not in mask):
                plt_mask = ax.imshow(slice_image_with_mask)
                plt_mask.set_visible(True)
            fig.canvas.mpl_connect('button_press_event', on_press)
            fig.canvas.draw()

        fig = plt.figure()
        position_list = [0, 0]
        plt.title(f'Slice {position_list[0]} channel {position_list[1]}')
        fig.canvas.mpl_connect('key_press_event', lambda event: key_event(event, position_list))
        ax = fig.add_subplot(111)
        slice_image = gray2rgb(chan_image[0])
        if masks is not None:
            colors = [(1, 1, 0), (0, 1, 1), (1, 0, 1)]
            for index, mask in enumerate(masks):
                slice_image = mark_boundaries(slice_image, mask[position_list[0]].astype(int), color=colors[index % len(colors)])
        ax.imshow(slice_image)

        lgr = logging.getLogger('Fuse')
        lgr.info('------------------------------------------')
        if metadata is not None:
            if isinstance(metadata, dict):
                lgr.info(FuseUtilsHierarchicalDict.to_string(metadata), {'color': 'magenta'})
            else:
                lgr.info(metadata)

        if label is not None and label != '':
            lgr.info('image label = ' + str(label), {'color': 'magenta'})
        lgr.info('------------------------------------------')

        plt.show()

    def visualize_aug(self, orig_sample: dict, aug_sample: dict, block: bool = True) -> None:
        """
        Visualise and compare augmented and non-augmented version of the sample
        :param orig_sample: batch_dict to extract the original sample from
        :param aug_sample: batch_dict to extract the augmented sample from
        :param block: set to False if the process should not be blocked until the plot will be closed
        :return: None
        """

        # extract data
        orig_image, orig_masks, orig_label, orig_metadata = self.extract_data(orig_sample)
        aug_image, aug_masks, aug_label, aug_metadata = self.extract_data(aug_sample)

        # visualize
        def key_event(e: Any, position_list: Any):  # using left/right key to move between slices
            def on_press(e: Any):  # use mouse click in order to toggle mask/no mask
                'toggle the visible state of the two images'
                if e.button:
                    # Toggle image with no augmentations
                    vis_image = plt_img.get_visible()
                    vis_mask = plt_mask.get_visible()
                    plt_img.set_visible(not vis_image)
                    plt_mask.set_visible(not vis_mask)
                    # Toggle image with augmentations
                    vis_aug_image = plt_aug_img.get_visible()
                    vis_aug_mask = plt_aug_mask.get_visible()
                    plt_aug_img.set_visible(not vis_aug_image)
                    plt_aug_mask.set_visible(not vis_aug_mask)

                    plt.draw()

            if e.key == "right":
                position_list[0] += 1
            elif e.key == "left":
                position_list[0] -= 1
            elif e.key == "up":
                position_list[1] += 1
            elif e.key == "down":
                position_list[1] -= 1
            else:
                return
            position_list[0] = position_list[0] % orig_image.shape[1]
            position_list[1] = position_list[1] % orig_image.shape[0]
            chan_image = orig_image[position_list[1]]
            chan_aug_image = aug_image[position_list[1]]
            # clearing subplots
            axs[0].cla()
            axs[1].cla()
            # creating image without augmentations and with toggling mask
            slice_image = gray2rgb(chan_image[position_list[0]])
            plt_img = axs[0].imshow(slice_image)
            plt_img.set_visible(False)

            if orig_masks is not None:
                slice_image_with_mask = slice_image
                colors = [(1, 1, 0), (0, 1, 1), (1, 0, 1)]
                for index, mask in enumerate(orig_masks):
                    slice_image_with_mask = mark_boundaries(slice_image_with_mask, mask[position_list[0]].astype(int),
                                                            color=colors[index % len(colors)])

                plt_mask = axs[0].imshow(slice_image_with_mask)
                plt_mask.set_visible(True)

            # creating image with augmentations and with toggling mask
            slice_aug_image = gray2rgb(chan_aug_image[position_list[0]])
            plt_aug_img = axs[1].imshow(slice_aug_image)
            plt_aug_img.set_visible(False)
            if aug_masks is not None:
                slice_aug_image_with_mask = slice_aug_image
                colors = [(1, 1, 0), (0, 1, 1), (1, 0, 1)]
                for index, mask in enumerate(aug_masks):
                    slice_aug_image_with_mask = mark_boundaries(slice_aug_image_with_mask, mask[position_list[0]].astype(int),
                                                                color=colors[index % len(colors)])

                plt_aug_mask = axs[1].imshow(slice_aug_image_with_mask)
                plt_aug_mask.set_visible(True)
            # drawing
            axs[0].title.set_text(f"Original - Slice {position_list[0]} channel {position_list[1]}")
            axs[1].title.set_text(f"Augmented - Slice {position_list[0]} channel {position_list[1]}")
            fig.canvas.mpl_connect('button_press_event', on_press)
            fig.canvas.draw()

        fig, axs = plt.subplots(ncols=2)
        position_list = [0, 0]
        chan_image = orig_image[position_list[1]]
        chan_aug_image = aug_image[position_list[1]]

        fig.canvas.mpl_connect('key_press_event', lambda event: key_event(event, position_list))
        slice_image = gray2rgb(chan_image[position_list[0]])
        if orig_masks is not None:
            colors = [(1, 1, 0), (0, 1, 1), (1, 0, 1)]
            for index, mask in enumerate(orig_masks):
                slice_image = mark_boundaries(slice_image, mask[position_list[0]].astype(int),
                                              color=colors[index % len(colors)])

        slice_aug_image = gray2rgb(chan_aug_image[0])
        if aug_masks is not None:
            colors = [(1, 1, 0), (0, 1, 1), (1, 0, 1)]
            for index, mask in enumerate(aug_masks):
                slice_aug_image = mark_boundaries(slice_aug_image, mask[position_list[0]].astype(int),
                                                  color=colors[index % len(colors)])

        axs[0].title.set_text(f"Original - Slice {position_list[0]} channel {position_list[1]}")
        axs[1].title.set_text(f"Augmented - Slice {position_list[0]} channel {position_list[1]}")
        axs[0].imshow(slice_image)
        axs[1].imshow(slice_aug_image)
        plt.show()



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import os
from typing import Dict
from fuse.managers.callbacks.callback_base import FuseCallback
from fuse.managers.manager_state import FuseManagerState
from fuse.utils.utils_file import FuseUtilsFile as file
from fuse.utils.utils_hierarchical_dict import FuseUtilsHierarchicalDict


class FuseTensorboardCallback(FuseCallback):
    """
        Responsible for writing the data of both training and validation to tensorborad loggers under model_dir.
    """

    def __init__(self, model_dir: str) -> None:
        super().__init__()
        self.model_dir = model_dir

    def on_step_end(self, step: int, train_results: Dict = None, validation_results: Dict = None, learning_rate: float = None) -> None:
        """
        Writes the results into train and validation loggers

        :param step: step number
        :param train_results: contains the data to be saved to train tensorboard.
                    e.g.: {'losses': {'loss1': mean_loss1,
                               'loss2': mean_loss2,
                                'total_loss': mean (loss1 + loss2)}
                    'metrics': {'metric1': epoch_metric1,
                                'metric2': epoch_metric2}}
        :param validation_results: contains the data to be saved to train tensorboard.
                    e.g.: {'losses': {'loss1': mean_loss1,
                               'loss2': mean_loss2,
                                'total_loss': mean (loss1 + loss2)}
                    'metrics': {'metric1': epoch_metric1,
                                'metric2': epoch_metric2}}
        :param learning_rate: the learning rate at the end of the step.
        """

        # update train tensorboard logger
        for evaluator_name in FuseUtilsHierarchicalDict.flatten(train_results).keys():
            evaluator_value = FuseUtilsHierarchicalDict.get(train_results, evaluator_name)
            if evaluator_value is not None and not isinstance(evaluator_value, (str,)):
                self.add_scalar(self.tensorboard_logger_train, tag=evaluator_name, scalar_value=evaluator_value, global_step=step)
        self.add_scalar(self.tensorboard_logger_train, 'learning_rate', learning_rate, step)

        # update train tensorboard logger
        if validation_results is not None:
            for evaluator_name in FuseUtilsHierarchicalDict.flatten(validation_results).keys():
                evaluator_value = FuseUtilsHierarchicalDict.get(validation_results, evaluator_name)
                if evaluator_value is not None and not isinstance(evaluator_value, (str,)):
                    self.add_scalar(self.tensorboard_logger_validation, tag=evaluator_name, scalar_value=evaluator_value, global_step=step)
            self.add_scalar(self.tensorboard_logger_validation, 'learning_rate', learning_rate, step)

        return

    def on_train_begin(self, state: FuseManagerState) -> None:
        """
        Called at the beginning of the train procedure.

        :param state: ignored
        """
        # File writer imports are done here in order to workaround the GPU issues -
        # when importing torch.tensorboard cuda gets occupied - do that only AFTER CUDA_VISIBLE_DEVICES is set
        try:
            # available only from torch 1.2
            from torch.utils.tensorboard import SummaryWriter
            self.writer_class = SummaryWriter
            self.use_summary_tf = False
        except ModuleNotFoundError:
            # fallback, use tensorflow file writer
            from tensorflow.summary import FileWriter
            import tensorflow as tf
            self.writer_class = FileWriter
            self.tf_summary = tf.Summary
            self.use_summary_tf = True

        tensorboard_train_dir = os.path.join(self.model_dir, 'train')
        tensorboard_validation_dir = os.path.join(self.model_dir, 'validation')

        # make sure we have these folders
        file.create_dir(tensorboard_train_dir, error_if_exist=False)
        file.create_dir(tensorboard_validation_dir, error_if_exist=False)

        # Get TensorBoard loggers
        self.tensorboard_logger_train = self.writer_class(tensorboard_train_dir)
        self.tensorboard_logger_validation = self.writer_class(tensorboard_validation_dir)
        pass

    def add_scalar(self, writer, tag, scalar_value, global_step):
        """Log a scalar variable."""
        if scalar_value is not None and not isinstance(scalar_value, str):
            if self.use_summary_tf:  # FileWrite
                summary = self.tf_summary(value=[self.tf_summary.Value(tag=tag, simple_value=scalar_value)])
                writer.add_summary(summary, global_step)
            else:
                writer.add_scalar(tag=tag, scalar_value=scalar_value, global_step=global_step)  # SummaryWriter


"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import logging
import os
from typing import Dict, List, Optional

import numpy as np
import pandas as pd
import torch
from torch import Tensor

from fuse.managers.callbacks.callback_base import FuseCallback
from fuse.utils.utils_file import FuseUtilsFile as file
from fuse.utils.utils_hierarchical_dict import FuseUtilsHierarchicalDict


class FuseInferResultsCallback(FuseCallback):
    """
        Responsible of writing the data of inference results into a CSV file.
        Collects the output data (corresponding to the output_columns) at the end of handle_batch into an aggregated dict,
            and writes it at the end of handle_epoch.
        The method self.get_infer_results() may be used to get the aggregated dict.
    """

    def __init__(self, output_file: Optional[str] = None, output_columns: Optional[List[str]] = None) -> None:
        super().__init__()
        self.output_columns = output_columns
        self.output_file = output_file

        # prepare output_path (if not already exists)
        if self.output_file is not None:
            file.create_dir(os.path.dirname(self.output_file))

        self.reset()
        pass

    def reset(self):
        self.aggregated_dict = {'descriptor': [], 'output': {}}
        self.infer_results_df = pd.DataFrame()

    def on_epoch_begin(self, mode: str, epoch: int) -> None:
        # make sure the aggregated dict is empty
        self.reset()
        pass

    def get_infer_results(self) -> pd.DataFrame:
        """
        Returns the aggregated results dict.
        :return: contains the key descriptor and a key for each metric/loss in output_columns.
        """

        return self.infer_results_df

    def on_epoch_end(self, mode: str, epoch: int, epoch_results: Dict = None) -> None:
        """
        When mode = 'infer' create a pickled DataFrame with inference results of self.output_columns.
        File is saved under self.output_path

        :param mode: write to file only on 'infer' mode.
        :param epoch: epoch number (ignored)
        :param epoch_results: not actually used
        """
        if mode != 'infer':
            return

        # prepare dataframe from the results
        infer_results_df = pd.DataFrame()
        infer_results_df['descriptor'] = self.aggregated_dict['descriptor']

        for output in FuseUtilsHierarchicalDict.get_all_keys(self.aggregated_dict['output']):
            infer_results_df[output] = list(
                FuseUtilsHierarchicalDict.get(self.aggregated_dict['output'], output))  # note- wrapping with list for pandas compatibility

        if self.output_file is not None:
            infer_results_df.to_pickle(self.output_file, compression='gzip')
            logging.getLogger('Fuse').info(f"Save inference results into {self.output_file}")

        self.reset()
        self.infer_results_df = infer_results_df
        return

    def on_batch_end(self, mode: str, batch: int, batch_dict: Dict = None) -> None:
        """
        On batch end - save the descriptor data into a class member.

        :param mode: accept only infer mode
        :param batch: batch number
        :param batch_dict: the batch input data
        """
        if mode != 'infer':
            return

        # for infer we need the descriptor and the output predictions
        descriptors = batch_dict['data'].get('descriptor', None)
        if isinstance(descriptors, Tensor):
            descriptors = list(descriptors.detach().cpu().numpy())
        self.aggregated_dict['descriptor'].extend(descriptors)

        if self.output_columns is not None and len(self.output_columns) > 0:
            output_cols = self.output_columns
        else:
            output_cols = FuseUtilsHierarchicalDict.get_all_keys(batch_dict)

        for output_col in output_cols:
            if output_col not in FuseUtilsHierarchicalDict.get_all_keys(self.aggregated_dict['output']):
                FuseUtilsHierarchicalDict.set(self.aggregated_dict['output'], output_col, [])
            output = FuseUtilsHierarchicalDict.get(batch_dict, output_col)
            if isinstance(output, torch.Tensor):
                # no need to save tensors
                output = output.cpu().numpy()
            FuseUtilsHierarchicalDict.get(self.aggregated_dict['output'], output_col).extend(output)

        pass



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Dict, List, Tuple, Callable, Optional, Sequence, Union
import numpy as np

from fuse.metrics.metric_base import FuseMetricBase
from fuse.metrics.metrics_toolbox import FuseMetricsToolBox


class FuseMetricConfusion(FuseMetricBase):
    """
    Multi class version for confusion metrics including: sensitivity, specificity, recall., precision, f1.

    """
    def __init__(self, pred_name: str, target_name: str, filter_func: Optional[Callable] = None, class_names:List = None,
                 operation_point:Optional[Union[float, List[Tuple], Callable]] = None,
                 metrics:Sequence[str] = ('sensitivity',), sum_weights:List = None, **kwargs):
        """
        :param pred_name:       batch_dict key for predicted output (e.g., class probabilities after softmax)
        :param target_name:     batch_dict key for target (e.g., ground truth label)
        :param filter_func: function that filters batch_dict/ The function gets ans input batch_dict and returns filtered batch_dict
        :param class_names:     Optional - name for each class otherwise the class index will be used.
        :param operation_point: Optional - few possible options:
                                    (1) None - the predicted class will be selected using argmax
                                    (2) float - binary classification, single threshold for the positive class/
                                    (3) threshold for each class, the order of the list is the order in which class
                                    is assigned, Each element in the list is a tuple of (class_idx, class_threshold)
                                    (4) Callable - getting as an input the prediction and targets and returns the format specified in (3)
        :param metrics: list of metrics name to return. See FuseMetricsToolBox.confusion_metrics for available options.
        :param sum_weights: weights for the weighted sum of sensitivity average

        """
        super().__init__(pred_name, target_name, filter_func, **kwargs)
        self._class_names = class_names
        self._operation_point = operation_point
        self._metrics = metrics
        self._sum_weights = sum_weights

    def process(self) -> Dict[str, float]:
        """
        Calculates the confusion metrics
        :return: dictionary including per class in one vs all manner and a macro avg for each class included in metrics

        """
        # find operation points
        if isinstance(self._operation_point, Callable):
            class_thresholds = self._operation_point(self.epoch_preds, self.epoch_targets)
        elif isinstance(self._operation_point, float):
            class_thresholds = [(1, self._operation_point), (0, 0.0)]
        else:
            class_thresholds = self._operation_point

        agg_results = {}

        # if no samples, return 'N/A'
        if len(self.epoch_preds) == 0:
            for metric_name in self._metrics:
                agg_results[metric_name + '_macro_avg'] = 'N/A'
            return agg_results

        # convert probabilities to class predictions
        epoch_class_preds = FuseMetricsToolBox.convert_probabilities_to_class(self.epoch_preds, class_thresholds)

        # set class names if not provided by the user
        class_names = FuseMetricsToolBox.get_class_names(self.epoch_preds[0].shape[-1], self._class_names)

        # set weight per class if not provided by the user
        if self._sum_weights is None:
            self._sum_weights = [1] * len(class_names)
        self._sum_weights = np.array(self._sum_weights)

        # compute the confusion metrics
        for cls_ind, cls in enumerate(class_names):
            cls_result = FuseMetricsToolBox.confusion_metrics(epoch_class_preds, cls_ind, np.array(self.epoch_targets), self._metrics)
            for metric_name in cls_result:
                if metric_name not in agg_results:
                    agg_results[metric_name] = {}
                metric_result = agg_results[metric_name]
                metric_result[cls] = cls_result[metric_name]

        # pack it in a dictionary and calculate macro average
        results = {}
        for metric_name in self._metrics:
            value_per_class = np.array(list(agg_results[metric_name].values()))
            indices = ~np.isnan(value_per_class)
            results.update({metric_name + '_' + cls: agg_results[metric_name][cls] for cls in agg_results[metric_name]})
            results[metric_name + '_macro_avg'] = np.average(value_per_class[indices], weights=self._sum_weights[indices])

        return results